#pragma once

#define _USE_MATH_DEFINES
#include <cmath>
#include <cassert>
#include <array>

// #define pi  M_PI
#define amu 1.66053904e-27 // kg
#define m_e 9.10938356E-31 // kg
#define k_B 1.38064852E-23 // J/K
#define landau 10

#define c_l 299792458 // m/s

#define q_e 1.6021766208E-19 // C
#define mu_0 4*M_PI*1e-7 // H/m
#define epsilon_0 1/(mu_0*c_l*c_l) // F/m

template<typename T>
T v_thermal_K(T temp_in_K, T mass = m_e){
    return std::sqrt(2/mass*temp_in_K*k_B);
}
template<typename T>
T v_thermal_eV(T temp_in_eV, T mass = m_e){
    return std::sqrt(2/mass*q_e*temp_in_eV);
}

template<typename T>
inline T maxwellian(T v, T v_thermal, T mu=0.){
    assert(v_thermal>0.);
    return std::exp(-std::pow((v-mu)/v_thermal,2))/(std::sqrt(M_PI)*v_thermal);
}
template<typename T, size_t dim>
inline T maxwellian(const std::array<T,dim>& v, T v_thermal, const std::array<T,dim>& mu = {}){
    assert(v_thermal>0.);
    T tmp_v = 0;
    for(size_t i=0; i<dim; ++i) {
        tmp_v+=std::pow(v[i]-mu[i],2);
    }
    return std::exp(-std::pow(tmp_v/v_thermal,2))/std::pow(std::sqrt(M_PI)*v_thermal,dim);
}
