#ifndef helpers_h
#define helpers_h

#include <sstream>
#include <vector>
#include <memory>

#include "optimize.h"

template<class ... Cs>
struct szeof;
template<class C, class ... Cs>
struct szeof<C, Cs...>{
    constexpr const static int value = 1 + szeof<Cs...>::value;
};
template<>
struct szeof<>{
    constexpr const static int value = 0;
};


template<int dim, typename T, typename I>
constexpr const I hash_array(const std::array<I,dim>& _in) {

    I out=0;
    I tmp=0;
    if (_in.size()>0){tmp=_in[0]*_in[0];}

    for(int i=1; i<dim; ++i){
        tmp = _in[i]-tmp;
        out *= (tmp+1) * _in[i];
    }

    return out;
}

const std::string random_string_lead = "RANDOM";
std::string random_string(){
    std::ostringstream output;
    output << random_string_lead;
    //NOTE: this is a solution, but could be improved.
    //     output+=std::to_string((std::chrono::time_point<std::chrono::system_clock,std::chrono::nanoseconds>::time_since_epoch()).count());
    //NOTE: a hash of the std::function would be much more suitable. But that does not exist.
    //     for(int i=0; i<8; ++i){
    output << std::rand()%(int)(1e6-1);
    //     }

    return output.str();
}


/// custom ostream operators

std::string array_separator = ",";
std::string array_element_head = "";
std::string array_element_tail = "";
template <class T, size_t N>
std::ostream& operator<<(std::ostream& o, const std::array<T, N>& arr)
{
    o << "{" ;
    for(auto iter = arr.cbegin(); iter!=arr.cend()-1; ++iter){
        o <<array_element_head<< *iter << array_element_tail<< array_separator;
    }
    o <<array_element_head<< arr.back() << array_element_tail<< "}";
    return o;
}
template <class T>
std::ostream& operator<<(std::ostream& o, const std::vector<T>& arr)
{
    if (arr.size()==0){
        o << "{}";
        return o;
    }
    o << "{" ;
    for(auto iter = arr.cbegin(); iter!=arr.cend()-1; ++iter){
        o <<array_element_head<< *iter << array_element_tail<< array_separator;
    }
    o <<array_element_head<< arr.back() << array_element_tail<< "}";
    return o;
}


template<typename T, typename I>
void inline swap_el(I pos1, I pos2, std::vector<T>& v){
    T tmp = v[pos2];
    v[pos2] = v[pos1];
    v[pos1] = tmp;
}

template<int dim, typename T, typename I>
void sort_v(std::vector<I>& lin, std::vector<T>& coeffi, std::vector<std::array<I,dim>>& pos){
    size_t i=0, j=1;
    I com;

    for(; i<lin.size(); ++i){
        j=i;
        com = lin[j];
        for(size_t k=j; k<lin.size(); ++k){
            if(lin[k]<com){
                com = lin[k];
                j=k;
            }
        }
        if(j>i){
            swap_el(i, j, lin);
            swap_el(i, j, coeffi);
            swap_el(i, j, pos);
        }
    }
}

template<typename type, size_t dim>
const constexpr std::array<type,dim> emplace_array(type val){
    std::array<type,dim> out {};
    out.fill(val);
    return out;
}

template<typename E>
auto enum_conversion(const E _value) -> typename std::underlying_type<E>::type {
    return static_cast<typename std::underlying_type<E>::type>(_value);
}


template<typename I, size_t s1, size_t s2>
std::array<I,s1+s2> array_concat(const std::array<I,s1>& _a1, const std::array<I,s2>& _a2){
    std::array<I,s1+s2> output {};
    std::copy(_a1.cbegin(),_a1.cend(),output.begin());
    std::copy(_a2.cbegin(),_a2.cend(),output.begin()+s1);

    return output;
}


template<typename T>
std::vector<T> operator+(const std::vector<T>& a1, const std::vector<T>& a2){
    assert(a1.size() == a2.size());
    std::vector<T> output(a1.size());
    for(size_t i=0; i<a1.size(); ++i){
        output[i]=a1[i]+a2[i];
    }
    return output;
}
template<typename T>
std::vector<T> operator-(const std::vector<T>& a1, const std::vector<T>& a2){
    assert(a1.size() == a2.size());
    std::vector<T> output(a1.size());
    for(size_t i=0; i<a1.size(); ++i){
        output[i]=a1[i]-a2[i];
    }
    return output;
}

template<typename T, size_t size>
std::array<T,size> operator+(const std::array<T,size>& a1, const std::array<T,size>& a2){
    std::array<T,size> output;
    for(size_t i=0; i<size; ++i){
        output[i]=a1[i]+a2[i];
    }
    return output;
}
template<typename T, size_t size>
std::array<T,size> operator-(const std::array<T,size>& a1, const std::array<T,size>& a2){
    std::array<T,size> output;
    for(size_t i=0; i<size; ++i){
        output[i]=a1[i]-a2[i];
    }
    return output;
}

template<typename T, size_t size>
std::array<T,size> operator+=(std::array<T,size>& a1, const std::array<T,size>& a2){
    for(size_t i=0; i<size; ++i){
        a1[i]+=a2[i];
    }
    return a1;
}

template<typename T>
std::vector<T> inline vector_shift(std::vector<T> vec, const T& val){
    for(auto& el: vec){
        el+=val;
    }
    return vec;
}

template<typename T>
void add_unique(const T& element, std::vector<T>& vector){
    for(auto& el: vector){
        if constexpr (std::is_arithmetic<T>::value){
            if(is_about_equal<T,T>(element,el)){
                return;
            }
        } else {
            if (element == el) {
                return;
            }
        }
    }
    vector.push_back(element);
}

template<bool B=true, typename T, size_t size>
std::array<T,size> apply_mask(const std::array<T,size>& ar, const std::array<bool,size>& m){
    std::array<T,size> output {};
    for(size_t i=0; i<size; ++i){
            output[i]=(m[i]==B)?ar[i]:0;
    }
    return output;
}

template<typename T>
std::vector<T> append_vector(std::vector<T>& v1, const std::vector<T>& v2){
    v1.insert(v1.end(),v2.cbegin(),v2.cend());
    return v1;
}

template <typename T> int sgn(T val) {
    return (T(0) < val) - (val < T(0));
}


#endif
