#pragma once

#include <cmath>
#include <vector>

#include "grid.h"


template<typename T>
struct L2 {
    constexpr static inline T direct(const T& v1, const T& v2) {
        return v1 * v2;
    }
    constexpr static inline T direct(const T& v1) {
        return v1*v1;
    }
    constexpr static inline T back(const T& v) {
        return std::sqrt(v);
    }
};


template<typename T, template <typename> class NORM=L2>
struct norm{

    constexpr static inline T direct(T val1, T val2){
        return NORM<T>::direct(val1, val2);
    }
    constexpr static inline T direct(T val1){
        return NORM<T>::direct(val1);
    }
    constexpr static inline T back(T val1){
        return NORM<T>::back(val1);
    }

    constexpr static inline T direct(const std::vector<T>& v1, const std::vector<T>& v2){
        T output = 0;
        assert(v1.size() == v2.size());
        auto iter1 = v1.cbegin();
        auto iter2 = v2.cbegin();
        for(; iter1!=v1.cend(); ++iter1, ++iter2){
            output+=direct(*iter1, *iter2);
        }
        return output;
    }

    constexpr static inline T self(const std::vector<T>& v1){
        T output = 0;
        auto iter1 = v1.cbegin();
        for(; iter1!=v1.cend(); ++iter1){
            output+=NORM<T>::direct(*iter1);
        }
        return NORM<T>::back(output);
    }

    template<typename I, class names, bool HD>
    constexpr static inline T direct(const grid<T,I,names,HD>& v1, const grid<T,I,names,HD>& v2){
        return direct(v1.storage,v2.storage);
    }

    template<typename I, class names, bool HD>
    constexpr static inline T self(const grid<T,I,names,HD>& v1){
        return self(v1.storage);
    }
};
