/**
 * @file optimize.h
 * @Author Markus Gasteiger
 * @date September, 2017
 * @brief These functions help to optimize the code. In that sense this file is related to debug.h
 *
 */

#ifndef optimize_h
#define optimize_h

#include <limits>
#include <chrono>
#include <cstdlib>
#include <type_traits>

#include <variant>

// as suggested at https://stackoverflow.com/questions/40447195/can-i-hint-the-optimizer-by-giving-the-range-of-an-integer
#if defined(NDEBUG)
#define assume(cond)
#else
#include <cassert>
#define assume(cond) assert(cond)
#endif

template<typename T, class c1>
bool inline is_about_zero(c1, T){
    return false;
}

template<typename T>
bool is_about_zero(T _input, T factor=10.){
    return std::abs(_input)<factor*std::numeric_limits<T>::epsilon();
}

template<typename T, class c1>
bool inline is_about_equal(c1 _i1, c1 _i2, T factor=10.){
    if constexpr (std::is_arithmetic<T>::value) {
        return is_about_zero<T,c1>(_i1-_i2,factor);
    } else {
        return _i1 == _i2;
    }
}


#endif
