#!/bin/python

from netCDF4 import Dataset as DS
import matplotlib.pyplot as plt
from numpy import log,vectorize
from math import *
import sys


m_e=9.10938356E-31
k_B=1.38064852E-23


def v_thermal_K(temp_in_k, mass = m_e):
    return sqrt(2/mass*temp_in_k*k_B)
def maxwellian(v, v_thermal, mu=0.):
    return exp(-((v-mu)/v_thermal)**2)/(sqrt(pi)*v_thermal);
vmaxwellian = vectorize(maxwellian)

def get_m_of_n(liste, m):
    from math import ceil

    pos=[]
    val=[]

    shift=(len(liste)-1)/(m-1)

    for i in range(m):
        pos.append(int(ceil(i*shift)))
        val.append(round(liste[pos[-1]],2))

    return pos,val

print("making cuts")
for fle in sys.argv[1:]:
    ds = DS(fle)
    try:
        plt.figure(figsize=(8,8))
        plt.imshow(ds['f'][:,:], aspect=1 )
        #extent=[ds['v'][0],ds['v'][-1],]
        #plt.xlim(ds['x'][0],ds['x'][-1])
        #plt.ylim(ds['v'][0],ds['v'][-1])
        pos,val = get_m_of_n(ds['x'][:],5)
        plt.xticks(pos,val)
        plt.xlabel('x')
        pos,val = get_m_of_n(ds['v'][:],5)
        plt.yticks(pos,val)
        plt.ylabel('$v/v_T$')

        plt.colorbar()
        plt.savefig(fle+'.pdf')
    except:
        import pdb; pdb.set_trace()
    plt.close()

    try:
        maxw=vmaxwellian(ds['v'][:],1)
        #import pdb; pdb.set_trace()
        plt.figure(figsize=(8,8))
        plt.plot(ds['v'],ds['f'][:,0],linestyle='-.', label="x=0%")
        plt.plot(ds['v'],ds['f'][:,int(ds['x'].size/4)],linestyle='--', label="x=25%")
        plt.plot(ds['v'],ds['f'][:,int(ds['x'].size/2)],linestyle='--', label="x=50%")
        plt.plot(ds['v'],ds['f'][:,int(ds['x'].size/4*3)],linestyle='--', label="x=75%")
        plt.plot(ds['v'],ds['f'][:,-1],linestyle='--', label="x=100%")
        plt.legend()
        plt.ylabel('f')
        plt.xlabel('$v/v_T$')
        #plt.plot(ds['v'],vmaxwellian(ds['v'][:],1))
        plt.savefig(fle+'_cut.pdf')
    except:
        import pdb; pdb.set_trace()
    plt.close()
