#ifndef axis_hpp
#define axis_hpp

#include "axis.h"


template<typename T, typename I, char NAME, discretization D, class ... f_args>
template<bool offset>
void axis<T,I,NAME,D,f_args...>::check_position(const I& _pos) const {
    if constexpr (offset) {
        assert(_pos >= -(I)gc_lower && "Position too small.");
        assert(_pos < (I)(points + gc_upper) && "Position too large.");
    } else {
        assert(_pos >= 0 && "Position too small.");
        assert(_pos < (I)(size) && "Position too large.");
    }
}

template<typename T, typename I, char NAME, discretization D, class ... f_args>
template<bool offset>
inline T axis<T,I,NAME,D,f_args...>::get_delta(I _pos) const {
    check_position<offset>(_pos);
    if constexpr (offset) {
        return internal_get_delta(_pos);
    } else {
        return internal_get_delta(_pos-(signed)this->gc_lower);
    }
}
template<typename T, typename I, char NAME, discretization D, class ... f_args>
template<bool offset>
inline T axis<T,I,NAME,D,f_args...>::get_distance(const I& _pos1, const I& _pos2) const {
    if constexpr (offset) {
        return internal_get_distance(_pos1, _pos2);
    } else {
        return internal_get_distance(_pos1-(signed)this->gc_lower, _pos2-(signed)this->gc_lower);
    }
}
template<typename T, typename I, char NAME, discretization D, class ... f_args>
inline T axis_hg<T,I,NAME,D,f_args...>::internal_get_distance(I _pos1, I _pos2) const {
//     this->template check_position<false>(_pos1);
//     this->template check_position<false>(_pos2);
    return this->delta * (_pos2-_pos1);
}
template<typename T, typename I, char NAME, discretization D, class ... f_args>
inline T axis_hg<T,I,NAME,D,f_args...>::internal_get_delta(I) const {
    return this->delta;
}

template<typename T, typename I, char NAME, discretization D, class ... f_args>
template<bool offset>
inline T axis<T,I,NAME,D,f_args...>::get_coord(I _pos) const {
    check_position<offset>(_pos);
    if constexpr (offset) {
        return internal_get_coord(_pos);
    } else {
        return internal_get_coord(_pos-(signed)this->gc_lower);
    }
}

template<typename T, typename I, char NAME, discretization D, class ... f_args>
inline T axis_hg<T,I,NAME,D,f_args...>::internal_get_coord(I _pos) const {
    return this->min+_pos*this->delta;
}

template<typename T, typename I, char NAME, discretization D, class ... f_args>
template<bool offset>
inline I axis<T,I,NAME,D,f_args...>::get_pos(const T& _value) const {
    I pos = internal_get_pos(_value);
    check_position<false>(pos);
    if constexpr (offset){
        pos-=this->gc_lower;
    }
    return pos;
}

template<typename T, typename I, char NAME, discretization D, class ... f_args>
inline size_t axis_hg<T,I,NAME,D,f_args...>::internal_get_pos(T _value) const {
    return std::round((_value-this->min)/this->delta)+this->gc_lower;
}
#endif
