#include <iostream>

#define BOOST_TEST_MODULE test_axis
#include <boost/test/unit_test.hpp>

#include "boundary.h"
#include "axis.h"
#include "debug.h"
#include "optimize.h"
#include "helpers.h"

using T = double; // less than double will make some of the tests invalid
using I = int;

BOOST_AUTO_TEST_SUITE( test_axis )

    boundary::boundary<T,I,std::array<T,3>> test1(boundary::btype::NEUMANN,1,[](const std::array<T,3>&){return 1.1;});
    //\TODO: The boundary functions currently have to be of the same type. This should be more general, i.e. the most general, suitable type should be used.
    boundary::boundary<T,I,std::array<T,3>> test2(boundary::btype::OPEN,0);
    boundary::boundary<T,I,std::array<T,3>> test3(boundary::btype::DIRICHLET,1,[](const std::array<T,3>&){return 1.1;});

BOOST_AUTO_TEST_CASE( initialization ){
    auto test_axis = make_axis<T,I,'x'>(0., 1., 101, test3, test1);
    auto test_axis2 = make_axis<T,I,'x'>(0., 1., 101, test1, test2);

    debug::msg2 << test_axis.print() << std::endl;
}

BOOST_AUTO_TEST_CASE( functions ){
    auto test_axis = make_axis<T,I,'x'>(0., 1.,101,test1,test2);
    debug::msg3 << (test_axis.get_delta(20) -0.01) << std::endl;
    BOOST_CHECK_EQUAL(test_axis.get_delta(20), 0.01);
    BOOST_CHECK_EQUAL(test_axis.get_coord(50),.5);
    BOOST_CHECK_EQUAL(test_axis.get_coord<false>(50),.49);
    debug::msg3 << test_axis.get_pos<false>(.5) << std::endl;
}

BOOST_AUTO_TEST_CASE( simplification ){
    auto test_axis = make_axis<T,I,'x'>(0., 1.,101,test3,test1);
    auto test_axis_simple = test_axis.simplify();
    BOOST_CHECK(test_axis_simple->cdscr == discretization::NONE);
    BOOST_CHECK(test_axis_simple->dscr == discretization::HOMOGENEOUS);
    BOOST_CHECK(test_axis.template get_coord<false>(2) == 0.01);
    BOOST_CHECK(test_axis_simple->template get_coord<false>(2) == 0.01);
}

BOOST_AUTO_TEST_SUITE_END()
