#pragma once

#include "grid.h"
#include "helpers.h"


template<typename T, typename I, class NAMES, bool HD>
template<bool offset, typename ITER>
ITER grid<T,I,NAMES,HD>::coords2pos(const std::array<I,dim>& in) const {
    static_assert(std::is_same<ITER,I>::value||std::is_same<ITER, typename std::vector<T>::iterator>::value);
    I out = props.template inner_product<offset>(in);
    assert(out>=0 && out< props.size);
    if constexpr (std::is_same<ITER,I>::value) { return out;}
    return storage.begin()+out;
}

template<typename T, typename I, class NAMES, bool HD>
template<bool offset, typename ITER>
std::array<I,NAMES::size> grid<T,I,NAMES,HD>::pos2coords(ITER in) const {
    static_assert(std::is_same<ITER,I>::value||std::is_same<ITER, typename std::vector<T>::iterator>::value);

    I local;
    if constexpr (std::is_same<ITER, typename std::vector<T>::iterator>::value) {
        local = (in-storage.begin());
    }
    assert(0<=local && local <props.size);
    return props.pos2coords<offset>(local);
}

template<typename T, typename I, class NAMES, bool HD>
template<bool offset, typename I2>
inline std::array<T,NAMES::size> grid<T,I,NAMES,HD>::get_deltas(const std::array<I2,dim>& ar) const{
    static_assert(std::numeric_limits<I2>::is_integer,"coords have to be integer");
    static_assert(std::is_signed<I2>::value || !offset, "the offset switch needs signed coords.");
    std::array<T,dim> out;
    for (size_t i=0; i<dim; ++i){
        out[i]=axes[i]->template get_delta<offset>(ar[i]);
    }
    return out;
}

template<typename T, typename I, class NAMES, bool HD>
template<bool offset, typename I2>
inline std::array<T,NAMES::size> grid<T,I,NAMES,HD>::get_coords(const std::array<I2,dim>& ar) const {
    static_assert(std::numeric_limits<I2>::is_integer,"coords have to be integer");
    static_assert(std::is_signed<I2>::value || !offset, "the offset switch needs signed coords.");
    std::array<T,dim> out;
    for(size_t i=0; i<dim; ++i){
        out[i]=axes[i]->template get_coord<offset>(ar[i]);
    }
    return out;
}

template<typename T, typename I, class NAMES, bool HD>
std::string grid<T,I,NAMES,HD>::print( const iterator<this_type*,basic_iterator>& range, const std::function<T(T)> _function) {
    std::ostringstream out;
    out.precision(2);

    // to adjust the shell tab width to be wide enough for the following output
    std::system("tabs 12"); // portability??

    std::string separator = "\t";

    int count=0;
    std::array<size_t,dim> new_beg;
    std::array<size_t,dim> new_end;
    std::array<bool,dim> shown_axis;
    std::vector<boundary::btype> used_boundaries {};

    std::string header="(Part of) grid for ";
    std::string label="";

    int one, two = -1;
    for (size_t ii=0; ii<dim; ++ii){
        shown_axis[ii] = true;
        if((range.end[ii]-range.begin[ii])>1 && count<2){
            new_end[ii] = range.end[ii];
            new_beg[ii] = range.begin[ii];
            count++;
            if (count==1) { one=ii;}
            else if (count==2) {two = ii;}
        } else {
            if ( props.limit_lower[ii]>=range.begin[ii] && props.limit_lower[ii] < range.end[ii]){
                new_beg[ii] = props.limit_lower[ii];
                new_end[ii] = props.limit_lower[ii]+1;
            } else {
                new_beg[ii] = range.begin[ii];
                new_end[ii] = range.begin[ii]+1;
            }
            header+=std::string(1,NAMES::list[ii])+"="+std::to_string(axes[ii]->template get_coord<false>(new_beg[ii]))+", ";
            shown_axis[ii] = false;
        }
    }

    if (count==2) {
        label="↓"+std::string(1,NAMES::list[two])+"\\"+label;
    }
    label+=std::string(1,NAMES::list[one])+"→";

    out << header << std::endl << label;
    for(size_t ii=range.begin[one]; ii<new_end[one]; ++ii){
        out << separator << axes[one]->template get_coord<false>(ii);
    }
    out << std::endl;
    auto itr = begin_intern<false,false>(new_beg,new_end);
    std::string tmp_out;
    size_t jj=0;
    if (count == 2) {
        jj=itr.begin[two];
    }
    T value = 0;
    do {
        if (itr.crds[one]==itr.begin[one]) {
            out << std::endl;
            if (count==2){
                out << axes[two]->template get_coord<false>(jj++);
            }
        }
        out << separator;
        out << itr.boundary_type();
        add_unique<boundary::btype>(itr.boundary_type(),used_boundaries);

        value = _function(*itr.ptr);
        if (is_about_zero(value)){
            out << 0.;
        } else {
//             if ( std::abs(*itr.ptr) >= 1000. || std::abs(*itr.ptr) <0.001)
            out << value;
        }
        out << boundary::btype::NONE;
//         out << "\033[0m";
    } while ((++itr).within());
    out << std::endl;
    out << std::endl;
    out << "Boundary types: ";
    for(auto& used_b: used_boundaries){
        if (used_b == boundary::btype::NONE){
            continue;
        }
        out << used_b << "🌑 " << boundary::name(used_b) << boundary::btype::NONE << "  ";
    }
    out << std::endl;
    out << "total grid size: "<< props.sizes << ", strides: " << props.strides;
    out << std::endl;

    return out.str();
}

template<typename T, typename I, class NAMES, bool HD>
std::string grid<T,I,NAMES,HD>::data_info(){
    std::ostringstream out;

    out << "Total dimension:\t" << props.sizes << std::endl;
    out << "Array size:\t\t" << props.size << std::endl;
    out << "Lower limits:\t\t" << props.limit_lower << std::endl;
    out << "Upper limits:\t\t" << props.limit_upper << std::endl;
    out << "Strides:\t\t" << props.strides << std::endl;

    return out.str();
}

template<typename T, typename I, class NAMES, bool HD>
template<bool offset, typename I2>
T& grid<T,I,NAMES,HD>::operator[](I2 _in) {
    if constexpr ( !offset ) {
        (signed)_in-=props.linear_lower_limit;
    }
    return storage[_in];
}

template<typename T, typename I, class NAMES, bool HD>
template<bool offset, typename I2>
T& grid<T,I,NAMES,HD>::operator[](const std::array<I2,dim>& _in) {
    I2 pos = props.template inner_product<offset>(_in);
    return storage[pos];
}

template<typename T, typename I, class NAMES, bool HD>
T& grid<T,I,NAMES,HD>::operator()(const std::array<T,dim>& _in){
    std::array<size_t,dim> pos {};
    for(size_t i=0; i<dim; ++i){
        pos[i] = axes[i]->template get_pos<false>(_in[i]);
    }
    return operator[](pos);
}

template<typename T, typename I, class NAMES, bool HD>
void grid<T,I,NAMES,HD>::copy_internal(const iterator<this_type*,basic_iterator>& source, const iterator<this_type*,basic_iterator>& target){
    assert(source.area()==target.area());
    do {
        // this is not good ... memcopy of continuous segments would be preferable
        *(target.ptr) = *(source.ptr);
    } while((++source).within() && (++target).within());
}

template<typename T, typename I, class NAMES, bool HD>
void grid<T,I,NAMES,HD>::copy_shift(iterator<this_type*,basic_iterator> target, std::make_signed<size_t>::type shift){
    for(; target.within(); ++target){
        *(target.ptr) = *(target.ptr+shift);
    }
}

template<typename T, typename I, class NAMES, bool HD>
void grid<T,I,NAMES,HD>::update_boundaries(){
    auto axis = axes.cbegin();
    size_t i = 0;
    for(; i<dim; ++i, ++axis){
        for(auto& is_upper: {false, true}){ // for parallelization
//             debug::msg2 << "o0-" << i << is_upper << name((*axis)->boundaries[is_upper].type) << std::endl;
            switch((*axis)->boundaries[is_upper].type){
                case boundary::btype::DIRICHLET: {
                    function((*axis)->boundaries[is_upper].function, begin_intern<false,true> (limits[i][0+is_upper], limits[i][2+is_upper]));
                }
                    break;
                case boundary::btype::NEUMANN:
                    derivative_on_range(i, is_upper, limits[i][0+is_upper], limits[i][2+is_upper], (*axis)->boundaries[is_upper].function);
                    break;
                case boundary::btype::OPEN:
                    break;
                case boundary::btype::PERIODIC:
                    copy_shift(begin_intern<basic_iterator> (limits[i][0+is_upper], limits[i][2+is_upper]), props.strides[i]*props.points[i]*(is_upper?-1:1));
                    break;
                case boundary::btype::MIRROR:
                    {
                        auto begin_l=limits[i][0+is_upper];
                        auto end_l=limits[i][2+is_upper];
                        end_l[i]=begin_l[i]+1;
                        size_t n = (limits[i][2+is_upper][i]-limits[i][0+is_upper][i]);
                        for(size_t j=0; j<n; ++j){
                            int lines = is_upper?-(1+2*j):(-1+2*(n-j));
                            copy_shift(begin_intern<basic_iterator> (begin_l, end_l), props.strides[i]*lines);
                            begin_l[i]++;
                            end_l[i]++;
                        }
                    }
                    break;
                case boundary::btype::NEIGHBOR:
                    assert(false && "Not implemented.");
                    break;
                default:
                    assert(false && "Boundary type not found.");
                    break;
            }
        }
    }
//     debug::msg2 << "o2" << std::endl;
}

template<typename T, typename I, class NAMES, bool HD>
void grid<T,I,NAMES,HD>::derivative_on_range(size_t axis, bool is_upper, const std::array<size_t,dim>& _begin, const std::array<size_t,dim>& _end, const std::function<T(const std::array<T,dim>&)>& _f_in) {
    assert(axis<NAMES::size);
    size_t thickness = _end[axis]-_begin[axis];
    assert(thickness>0);


    std::function<T(T*, const std::array<T,dim>&)> f_local;
    std::make_signed<size_t>::type linear_offset = 0;
    T delta = 0.;

    std::array<size_t,dim> begin_l=_begin, end_l=_end;
    end_l[axis]=begin_l[axis]+1;

//     std::string tstr = is_upper?"(upper)":"(lower)";
//     debug::msg2 << "axis" << tstr << axis << ": " << _begin << "-" << _end << std::endl;
    for(size_t i=_begin[axis]; i<_end[axis]; ++i){
        linear_offset = props.strides[axis];
        if(is_upper){
            delta = axes[axis]->template get_distance<false>(props.limit_upper[axis]-1,i);
            debug::msg3 << props.limit_upper[axis]-1 << " " << i << " -> "<< delta<< std::endl;
            assert(delta>0.);

            linear_offset *= (props.limit_upper[axis]-1 - i);
        } else {
            delta = axes[axis]->template get_distance<false>(props.limit_lower[axis],i);
            assert(delta<0.);

            debug::msg3 << i << "-" << props.limit_lower[axis] << std::endl;
            linear_offset *= (props.limit_lower[axis] - i);
        }

        f_local = [=](T* val, const std::array<T,dim>& _ar)->T {
            std::array<T,dim> tmp = _ar;
            tmp[axis] = -delta;
            T base = *(val+linear_offset);
            T result = base + delta*_f_in(tmp);
            return result;
        };
        begin_l[axis]=i;
        end_l[axis]=i+1;

        function<std::function<T(T*, const std::array<T,dim>&)>,true,this_type*,false,iter_type<false,true>::delta_coords>(f_local, begin_intern<iter_type<false,true>::delta_coords>(begin_l,end_l));

//         debug::level = tmpll;
    }
}

template<typename T, typename I, class NAMES, bool HD>
auto grid<T,I,NAMES,HD>::set_limits(){
    std::array<std::array<std::array<size_t,NAMES::size>,4>,NAMES::size> output {};
    std::array<size_t,NAMES::size> zeroes {};

    for(size_t i=0; i<NAMES::size; ++i){
        output[i][0]=zeroes;

        output[i][1]   =zeroes;
        output[i][1][i]=props.limit_upper[i];

        output[i][2]   =props.sizes;
        output[i][2][i]=props.limit_lower[i];

        output[i][3]=props.sizes;
    }

    return output;
}

template<typename T, typename I, class NAMES, bool HD>
void grid<T,I,NAMES,HD>::change_boundaries(const char _b, const bool is_upper, const boundary::boundary<T,I,std::array<T,dim>>& _new){
    size_t pos=0;
    for(; pos<NAMES::size; ++pos){
        if (_b == NAMES::list[pos]){
            break;
        }
    }
    assert(pos<NAMES::size && "This grid does not contain the given dimension.");

    assert(_new.gc<=axes[pos]->boundaries[is_upper].gc && "The new boundary must not be thicker than the previous one.");

    (axes[pos]->boundaries[is_upper])=_new;
}
