#pragma once

#include "grid.h"

/// reduction part

template<typename T, typename I, class NAMES, bool HD>
template<class REDUCTION,
        typename FUNCTION,
        bool iter_direct,
        template <typename,bool,typename,typename,class> class ITER_TYPE>
void grid<T,I,NAMES,HD>::reduction_function(REDUCTION& R,
                                            const FUNCTION& _f,
                                            iterator<this_type*,ITER_TYPE> range){
    constexpr bool use_deltas = ITER_TYPE<this_type*,HD,T,I,NAMES>::uses_deltas;
    constexpr bool use_coords = ITER_TYPE<this_type*,HD,T,I,NAMES>::uses_coords;

    for(; range.within(); ++range){
        assert((range.ptr-&storage[0])>=0);
        assert((range.ptr-&storage[0]) < (signed) storage.size());
        if constexpr (iter_direct) {
            R.append(_f(range));
        } else {
            if constexpr (use_deltas && use_coords) {
                R.append(_f(*range.ptr, range.deltas, range.coords));
            } else if constexpr (use_deltas && !use_coords) {
                R.append(_f(*range.ptr, range.deltas));
            } else if constexpr (!use_deltas && use_coords) {
                R.append(_f(*range.ptr, range.coords));
            } else if constexpr (!use_deltas && !use_coords) {
                R.append(_f(*range.ptr));
            }
        }
    }
}


template<typename T, typename I, class NAMES, bool HD>
template<template <typename,typename> class R,
        class new_names,
        typename FUNCTION,
        bool iter_direct,
        template <typename,bool,typename,typename,class> class ITER_TYPE>
auto grid<T,I,NAMES,HD>::reduction(const FUNCTION& _f){
    auto mask =NAMES::template mask<new_names>();
    auto end = apply_mask(props.limit_upper,mask);
    auto shift = apply_mask<false>(props.limit_upper-props.limit_lower,mask);
    auto output = this->template sub_grid<new_names>();

    R<T,I> reducer;
    auto out = output.template begin_intern<basic_iterator>(output.props.limit_lower,output.props.limit_upper);

    for(auto ranger = begin_intern<basic_iterator>(props.limit_lower,end); ranger.within(); ++ranger, ++out){
        reducer.reset();
        reduction_function<R<T,I>,FUNCTION,iter_direct,ITER_TYPE>(reducer, _f, begin_intern<ITER_TYPE>(ranger.crds,ranger.crds+shift));

        *out.ptr = reducer.get();
    }

    return output;
}

template<typename T, typename I, class NAMES, bool HD>
template<template <typename,typename> class R,
        class new_names,
        bool use_deltas, bool use_coords,
        typename FUNCTION>
auto grid<T,I,NAMES,HD>::reduction(const FUNCTION& _f){
    return reduction<R,new_names,FUNCTION,false,iter_type<use_deltas,use_coords>::template delta_coords>(_f);
}

template<typename T, typename I, class NAMES, bool HD>
template<template <typename,typename> class R,
        class new_names,
        template <typename,bool,typename,typename,class> class ITER_TYPE>
auto grid<T,I,NAMES,HD>::reduction(const std::function<T(const iterator<this_type*,ITER_TYPE>&)>& _f){
    return reduction<R,new_names,std::function<T(const  iterator<this_type*,ITER_TYPE>&)>,true,ITER_TYPE>(_f);
}


//// function part

template<typename T, typename I, class NAMES, bool HD>
template<typename FUNCTION,
        bool access,
        typename POINTER,
        bool iter_direct,
        template <typename,bool,typename,typename,class> class ITER_TYPE>
void grid<T,I,NAMES,HD>::function(const FUNCTION& _f, iterator<POINTER,ITER_TYPE> range){
    constexpr bool use_deltas = ITER_TYPE<POINTER,HD,T,I,NAMES>::uses_deltas;
    constexpr bool use_coords = ITER_TYPE<POINTER,HD,T,I,NAMES>::uses_coords;

    for(; range.within(); ++range){
        assert((range.ptr-&storage[0])>=0);
        assert((range.ptr-&storage[0]) < (signed) storage.size());
        if constexpr (iter_direct) {
            *(range.ptr) = _f(range);
        } else {
            if constexpr (access && use_deltas && use_coords) {
                *(range.ptr) = _f(range.ptr, range.deltas, range.coords);
            } else if constexpr (access && use_deltas && !use_coords) {
                *(range.ptr) = _f(range.ptr, range.deltas);
            } else if constexpr (access && !use_deltas && use_coords) {
                *(range.ptr) = _f(range.ptr, range.coords);
            } else if constexpr (access && !use_deltas && !use_coords) {
                *(range.ptr) = _f(range.ptr);
            } else if constexpr (!access && use_deltas && use_coords) {
                *(range.ptr) = _f(range.deltas, range.coords);
            } else if constexpr (!access && use_deltas && !use_coords) {
                *(range.ptr) = _f(range.deltas);
            } else if constexpr (!access && !use_deltas && use_coords) {
                *(range.ptr) = _f(range.coords);
            } else if constexpr (!access && !use_deltas && !use_coords) {
                *(range.ptr) = _f();
            }
        }
    }
}


//NOTE: is there any chance of simplification?
template<typename T, typename I, class NAMES, bool HD>
template<typename FUNCTION,
        bool access,
        typename POINTER,
        bool iter_direct,
        template <typename,bool,typename,typename,class> class ITER_TYPE>
void grid<T,I,NAMES,HD>::function(const FUNCTION& _f, iterator<POINTER,ITER_TYPE> range, grid<T,I,NAMES,HD>& output){
    constexpr bool use_deltas = ITER_TYPE<this_type*,HD,T,I,NAMES>::uses_deltas;
    constexpr bool use_coords = ITER_TYPE<this_type*,HD,T,I,NAMES>::uses_coords;

    for(; range.within(); ++range) {
        assert((range.ptr-&storage[0])<storage.size());
        if constexpr (iter_direct) {
                output(range.ptr-storage.data()) = _f(range);
        } else {
            if constexpr (access && use_deltas && use_coords) {
                output(range.ptr-storage.data()) = _f(range.ptr, range.deltas, range.coords);
            } else if constexpr (access && use_deltas && !use_coords) {
                output(range.ptr-storage.data()) = _f(range.ptr, range.deltas);
            } else if constexpr (access && !use_deltas && use_coords) {
                output(range.ptr-storage.data()) = _f(range.ptr, range.coords);
            } else if constexpr (access && !use_deltas && !use_coords) {
                output(range.ptr-storage.data()) = _f(range.ptr);
            } else if constexpr (!access && use_deltas && use_coords) {
                output(range.ptr-storage.data()) = _f(range.deltas, range.coords);
            } else if constexpr (!access && use_deltas && !use_coords) {
                output(range.ptr-storage.data()) = _f(range.deltas);
            } else if constexpr (!access && !use_deltas && use_coords) {
                output(range.ptr-storage.data()) = _f(range.coords);
            } else if constexpr (!access && !use_deltas && !use_coords) {
                output(range.ptr-storage.data()) = _f();
            }
        }
    }
}


template<typename T, typename I, class NAMES, bool HD>
template<bool access,
        template <typename,bool,typename,typename,class> class ITER_TYPE>
void grid<T,I,NAMES,HD>::function(const typename function_type_e<NAMES::size,T,I,T*,access,ITER_TYPE<this_type*,HD,T,I,NAMES>>::type& _f){
    function<typename function_type_e<NAMES::size,T,I,T*,access,ITER_TYPE<this_type*,HD,T,I,NAMES>>::type, access, this_type*, false, ITER_TYPE>(_f, begin_intern<ITER_TYPE>(props.limit_lower,props.limit_upper
    ));

}
//TODO: merge these functions, they are redundant
template<typename T, typename I, class NAMES, bool HD>
template<bool access, bool use_deltas, bool use_coords>
void grid<T,I,NAMES,HD>::function(const typename function_type<NAMES::size,T,I,T*,access,use_deltas,use_coords>::type& _f){
    function<access, iter_type<use_deltas,use_coords>::template delta_coords>(_f);
    update_boundaries();
}

template<typename T, typename I, class NAMES, bool HD>
template<template <typename,bool,typename,typename,class> class ITER_TYPE>
void grid<T,I,NAMES,HD>::function(const std::function<T(const iterator<this_type*,ITER_TYPE>&)>& _f){
    function<std::function<T(const iterator<this_type*,ITER_TYPE>&)>, false, this_type*, true, ITER_TYPE>(_f, begin_intern<ITER_TYPE>(props.limit_lower,props.limit_upper));
    update_boundaries();
}
