#ifndef grid_iterator_h
#define grid_iterator_h

#include "grid.h"
#include "functional_types.h"
#include "grid_iterator_delta_coords.h"

#include <cmath>
#include <memory>
#include <array>
#include <iterator>
#include <type_traits>

// template<typename,bool,typename,typename,class>
// using ITER_TYPE = grid_iterator_delta_coords;


template<typename T, typename I, class NAMES, bool HD>
template<typename POINTER, template <typename, bool, typename, typename, class> class ITER_TYPE>
struct grid<T,I,NAMES,HD>::iterator : public ITER_TYPE<POINTER, HD, T, I, NAMES>  {
    using std_function_type = typename function_type<NAMES::size, T, I, T, true, ITER_TYPE<POINTER,HD,T,I,NAMES>::uses_deltas, ITER_TYPE<POINTER,HD,T,I,NAMES>::uses_coords>::type;
    using ITER_TYPE<POINTER, HD,T,I,NAMES>::parent;
    using ITER_TYPE<POINTER, HD,T,I,NAMES>::crds;
    using ITER_TYPE<POINTER, HD,T,I,NAMES>::deltas;
    using ITER_TYPE<POINTER, HD,T,I,NAMES>::coords;

//     iterator(const std::array<size_t,dim>& _begin, const std::array<size_t,dim>& _end) : begin(_begin), end(_end) {};
    iterator(POINTER _parent, const std::array<size_t,dim>& _begin, const std::array<size_t,dim>& _end) : ITER_TYPE<POINTER,HD,T,I,NAMES>(_parent,_begin), ptr_0(&((_parent->storage)[0])), begin(_begin), end(_end) {
        static_assert(std::is_same<POINTER,std::shared_ptr<grid<T,I,NAMES,HD>>>::value || std::is_same<POINTER,grid<T,I,NAMES,HD>*>::value);

        ptr= &ptr_0[parent->props.template inner_product<false>(begin)];
        this->set();
    };

    ~iterator() {
        ptr=nullptr;
        ptr_0=nullptr;
        if constexpr (std::is_same<POINTER,std::shared_ptr<grid<T,I,NAMES,HD>>>::value){
            parent.reset();
        } else if constexpr (std::is_same<POINTER,grid<T,I,NAMES,HD>*>::value) {
            parent=nullptr;
        }
    }

    T* ptr;
    T* ptr_0;
    const std::array<size_t,dim> begin;
    const std::array<size_t,dim> end;

    auto& prev();
    auto& next();
    auto& operator++();
    auto& operator--();

    auto& operator+(const std::array<size_t,dim>&);
    auto& operator+=(const std::array<size_t,dim>&);

    void reset();
    void reset_end();

//     template<bool ud, bool uc>
//     bool operator!=(const iterator<ud,uc>& _in) const;
    bool within() const;
    size_t area() const;
    std::array<size_t,NAMES::size> thickness() const;

    bool is_boundary() const;
    boundary::btype boundary_type() const;

    T& operator*() { return *ptr; }
    T& operator*() const { return *ptr; }

    void validate();
    size_t lin_pos();
};

#include "grid_iterator.hpp"
#endif
