#ifndef grid_iterator_hpp
#define grid_iterator_hpp

#include "grid_iterator.h"

template<typename T, typename I, class NAMES, bool HD>
template<template <typename, bool, typename, typename, class> class ITER_TYPE>
inline auto grid<T,I,NAMES,HD>::begin(const std::array<size_t,dim>& begin, const std::array<size_t,dim>& end) {
    return grid<T,I,NAMES,HD>::iterator<std::shared_ptr<grid<T,I,NAMES,HD>>,ITER_TYPE> (this->shared_from_this(), begin, end);
}

// template<typename T, typename I, class NAMES, bool HD>
// template<template <typename, bool, typename, typename, class> class ITER_TYPE>
// inline auto grid<T,I,NAMES,HD>::begin_intern(const std::array<size_t,dim>& begin, const std::array<size_t,dim>& end) {
//     return grid<T,I,NAMES,HD>::iterator<this_type*, ITER_TYPE> (this,begin,end);
// }

template<typename T, typename I, class NAMES, bool HD>
template<typename POINTER, template <typename, bool, typename, typename, class> class ITER_TYPE>
auto& grid<T,I,NAMES,HD>::iterator<POINTER,ITER_TYPE>::next() {
    size_t i=0;
    for(; i<NAMES::size; ++i){
        crds[i]++;
        if(crds[i] < end[i]){
            ptr+=this->parent->props.strides[i];

            this->add(i);
            goto endn;
        } else {
            ptr+=(begin[i]-crds[i]+1)* parent->props.strides[i];
            crds[i]=begin[i];

            this->set(i);
        }
    }

    ptr=nullptr;
    crds=end;

    endn:
    return *this;
}

template<typename T, typename I, class NAMES, bool HD>
template<typename POINTER, template <typename, bool, typename, typename, class> class ITER_TYPE>
auto& grid<T,I,NAMES,HD>::iterator<POINTER,ITER_TYPE>::prev() {
    size_t i=0;
    for(; i<NAMES::size; ++i){
        if(crds[i] > begin[i]){
            crds[i]--;
            ptr-=parent->props.strides[i];

            this->subtract(i);
            goto endn;
        } else {
            ptr+=(end[i]-crds[i]-1)* parent->props.strides[i];
            crds[i]=end[i]-1;

            this->set(i);
        }
    }

    ptr=nullptr;
    crds=begin;

    endn:
    return *this;
}

template<typename T, typename I, class NAMES, bool HD>
template<typename POINTER, template <typename, bool, typename, typename, class> class ITER_TYPE>
auto& grid<T,I,NAMES,HD>::iterator<POINTER,ITER_TYPE>::operator++() {
    return this->next();
}

template<typename T, typename I, class NAMES, bool HD>
template<typename POINTER, template <typename, bool, typename, typename, class> class ITER_TYPE>
auto& grid<T,I,NAMES,HD>::iterator<POINTER,ITER_TYPE>::operator--() {
    return this->prev();
}

template<typename T, typename I, class NAMES, bool HD>
template<typename POINTER, template <typename, bool, typename, typename, class> class ITER_TYPE>
auto& grid<T,I,NAMES,HD>::iterator<POINTER,ITER_TYPE>::operator+=(const std::array<size_t,dim>& _array) {
    size_t i=0;
    using ssize_t = std::make_signed<size_t>::type;
    ssize_t overlap=0;
    ssize_t tmp=0;

    for(; i<NAMES::size; ++i){
        crds[i]+=_array[i]+overlap;

        if(crds[i] < end[i] && crds[i]>=begin[i]){
            ptr+=(_array[i]+overlap)* parent->props.strides[i];

            this->add(i, _array[i]+overlap);

            overlap=0;
        } else {
            if (_array[i]+overlap <0){
                tmp=end[i];
            } else {
                tmp=begin[i]-1;
            }
            tmp+=(ssize_t) crds[i]%(end[i]-begin[i]);
            overlap=std::floor(crds[i]/(end[i]-begin[i]));
            crds[i]=tmp;

            ptr+=(begin[i]-crds[i]+1)* parent->props.strides[i];

            this->set(i);
        }
    }

    if (overlap!=0){
        ptr=nullptr;
        crds=end;
    }

    return *this;
}

template<typename T, typename I, class NAMES, bool HD>
template<typename POINTER, template <typename, bool, typename, typename, class> class ITER_TYPE>
bool grid<T,I,NAMES,HD>::iterator<POINTER,ITER_TYPE>::within() const {
//     debug::msg2 << "m2" << (ptr==nullptr) <<std::endl;
    return ptr!=nullptr;
//     if(ptr==nullptr){
//         return false;
//     }
//     return true;
}
template<typename T, typename I, class NAMES, bool HD>
template<typename POINTER, template <typename, bool, typename, typename, class> class ITER_TYPE>
size_t grid<T,I,NAMES,HD>::iterator<POINTER,ITER_TYPE>::area() const {
    size_t output=1;
    for(size_t i=0; i<NAMES::size; ++i){
        output*=(end[i]-begin[i]);
    }
    return output;
}

template<typename T, typename I, class NAMES, bool HD>
template<typename POINTER, template <typename, bool, typename, typename, class> class ITER_TYPE>
void grid<T,I,NAMES,HD>::iterator<POINTER,ITER_TYPE>::validate(){
    assert(ptr>=parent->storage.data());
    assert(ptr< parent->storage.data()+parent->props.size);
}

template<typename T, typename I, class NAMES, bool HD>
template<typename POINTER, template <typename, bool, typename, typename, class> class ITER_TYPE>
size_t grid<T,I,NAMES,HD>::iterator<POINTER,ITER_TYPE>::lin_pos(){
    return ptr - ptr_0;
}

template<typename T, typename I, class NAMES, bool HD>
template<typename POINTER, template <typename, bool, typename, typename, class> class ITER_TYPE>
std::array<size_t,NAMES::size> grid<T,I,NAMES,HD>::iterator<POINTER,ITER_TYPE>::thickness() const {
    return end-begin;
}
template<typename T, typename I, class NAMES, bool HD>
template<typename POINTER, template <typename, bool, typename, typename, class> class ITER_TYPE>
boundary::btype grid<T,I,NAMES,HD>::iterator<POINTER,ITER_TYPE>::boundary_type() const {
    for(size_t i=0; i<dim; ++i){
        if ( crds[i] < parent->props.limit_lower[i] ){
            return parent->axes[i]->boundaries[0].type;
        } else if (crds[i] >= parent->props.limit_upper[i]) {
            return parent->axes[i]->boundaries[1].type;
        }
    }
    return boundary::btype::NONE;
}
template<typename T, typename I, class NAMES, bool HD>
template<typename POINTER, template <typename, bool, typename, typename, class> class ITER_TYPE>
bool grid<T,I,NAMES,HD>::iterator<POINTER,ITER_TYPE>::is_boundary() const {
    for(size_t i=0; i<dim; ++i){
        if ( crds[i] < parent->props.limit_lower[i] || crds[i] >= parent->props.limit_upper[i]) { return true;}
    }
    return false;
}

template<typename T, typename I, class NAMES, bool HD>
template<typename POINTER, template <typename, bool, typename, typename, class> class ITER_TYPE>
void grid<T,I,NAMES,HD>::iterator<POINTER,ITER_TYPE>::reset(){
    ptr = &ptr_0[parent->props.inner_product(begin)];
    crds = begin;
    this->set();
    coords = parent->template get_coords<false>(begin);
}

template<typename T, typename I, class NAMES, bool HD>
template<typename POINTER, template <typename, bool, typename, typename, class> class ITER_TYPE>
void grid<T,I,NAMES,HD>::iterator<POINTER,ITER_TYPE>::reset_end(){
    crds = end;
    for(auto& el: crds) { el--;}
    this->set();
    ptr = &ptr_0[parent->props.template inner_product<false>(crds)];
    coords = parent->template get_coords<false>(crds);
}
#endif
