#pragma once

template<bool use_deltas, bool use_coords, typename POINTER, bool HD, typename T, typename I, class NAMES>
struct delta_coords{
    constexpr static bool uses_coords = use_coords;
    constexpr static bool uses_deltas = use_deltas;

    delta_coords(POINTER _parent, const std::array<size_t,NAMES::size>& begin) : parent(_parent), crds(begin) {
        deltas = parent->template get_deltas<false>(begin);
        coords = parent->template get_coords<false>(begin);
    }

    POINTER parent;
    std::array<size_t,NAMES::size> crds;
    std::array<T,NAMES::size> deltas;
    std::array<T,NAMES::size> coords;

    inline void add(size_t i, size_t shift = 1){
        if constexpr((use_deltas || use_coords)&&!HD){
            deltas[i]=parent->axes[i].template get_delta<false>(crds[i]);
        }
        if constexpr(use_coords && HD){
            coords[i]+=shift*deltas[i];
        } else if constexpr (use_coords && !HD){
            coords[i]=parent->axes[i]->template get_coord<false>(crds[i]);
        }
    }

    inline void subtract(size_t i){
        if constexpr((use_deltas || use_coords)&&!HD){
            deltas[i]=parent->axes[i].get_delta(crds[i]);
        }
        if constexpr(use_coords && HD){
            coords[i]-=deltas[i];
        } else if constexpr (use_coords && !HD){
            coords[i]=parent->axes[i]->template get_coord<false>(crds[i]);
        }
    }

    inline void set(size_t i){
        if constexpr(use_deltas && !HD){
            deltas[i]=parent->axis[i]->template get_delta<false>(crds[i]);
        }
        if constexpr(use_coords){
            coords[i]=parent->axes[i]->template get_coord<false>(crds[i]);
        }
    }

    inline void set(){
        if constexpr(use_deltas && !HD){
            deltas=parent->template get_deltas<false>(crds);
        }
        if constexpr(use_coords){
            coords=parent->template get_coords<false>(crds);
        }
    }
};

template<typename POINTER, bool HD, typename T, typename I, class NAMES>
using basic_iterator = delta_coords<false,false, POINTER, HD, T, I, NAMES>;

template<bool _ud, bool _uc, typename _pointer, template <bool, bool, typename, bool, typename, typename, class> class _ITER>
struct rest_iter{
    template<typename _P, bool _HD, typename _T, typename _I, class _Names>
    using tmp_type = _ITER<_ud,_uc,_P,_HD,_T,_I,_Names>;
//     using type = iterator<_pointer, tmp_type>;
};


template<bool use_deltas, bool use_coords>
struct iter_type{
    template<typename _P, bool _HD, typename _T, typename _I, class _Names>
    using delta_coords = delta_coords<use_deltas,use_coords,_P,_HD,_T,_I,_Names>;
};
