#include <iostream>

#define BOOST_TEST_MODULE test_axis
#include <boost/test/unit_test.hpp>

#include <string>
#include <array>

#include "grid.h"
#include "debug.h"
#include "optimize.h"
#include "export.h"
#include "properties.h"
#include "axis.h"
#include "reduction.h"

#include "grid_iterator_delta_coords.h"

using T = double; // less than double will make some of the tests invalid
using I = int;

BOOST_AUTO_TEST_SUITE( test_grid )

template<typename T>
inline T testfunction(const std::array<T,3>& ar){
    return ar[1]+ar[0]/10.+ar[2]/100.;
}

// boundaries
boundary::boundary<T,I,std::array<T,3>> test1a(boundary::btype::NEUMANN,2,[](const std::array<T,3>&){ return 2.11;});
boundary::boundary<T,I,std::array<T,3>> test1(boundary::btype::NEUMANN,1,[](const std::array<T,3>&){ return 1.11;});
//\TODO: The boundary functions currently have to be of the same type. This should be more general, i.e. the most general, suitable type should be used.
boundary::boundary<T,I,std::array<T,3>> test2(boundary::btype::OPEN,0);
boundary::boundary<T,I,std::array<T,3>> test3(boundary::btype::DIRICHLET,1,[](const std::array<T,3>&){ return 1.13;});
boundary::boundary<T,I,std::array<T,3>> test4(boundary::btype::MIRROR,2);
boundary::boundary<T,I,std::array<T,3>> test5(boundary::btype::PERIODIC,2);
    std::array<size_t,3> ttmp = {1,1,1};

// axes
auto axis_x = make_axis<T,I,'x'>(0., 1.,101,test3,test1);
auto axis_y = make_axis<T,I,'y'>(0., 1.,104,test1,test2);
auto axis_z = make_axis<T,I,'z'>(0., 1.,107,test2,test3);

BOOST_AUTO_TEST_CASE ( test_axes2type ){
// debug::level = debug_level::FULL;
    using ttest = liste<I,1,2,6>;
    ttest sttest = liste<I,1,2,6>();
    BOOST_CHECK_EQUAL(sttest.inner_product(std::array<I,3>({1,1,1})),9);
    BOOST_CHECK_EQUAL((std::array<I,3>{{1,1,1}}*sttest),9);
    BOOST_CHECK((9%sttest)==(std::array<I,3>{{1,1,1}}));

    using ttest2 = liste<I,3,4,8>;
    using MAX = liste<long unsigned int, 10, 11, 12>;
    using list_sum = typename lists_add<I,ttest,ttest2>::reslt;
    BOOST_CHECK_EQUAL(list_sum::template value<0>(),4);
    BOOST_CHECK_EQUAL((listval<0,I,list_sum>::val),4);
    BOOST_CHECK(0 < (listval<0,I,list_sum>::val));
    BOOST_CHECK_EQUAL(MAX::template value<1>(),11);
}

BOOST_AUTO_TEST_CASE ( test_properties ){
    auto test_properties = properties<T,I,dim<'a','b','c'>>();
    test_properties.template inner_product(ttmp);

}

auto axis_1 = make_axis<T,I,'x'>(0., 1.,4,test3,test1a);
auto axis_2 = make_axis<T,I,'y'>(0., 1.,6,test1,test2);
auto axis_3 = make_axis<T,I,'z'>(0., 1.,4,test4,test5);
BOOST_AUTO_TEST_CASE ( grid_initialization ){
    auto gitter1 = make_grid<T,I,true>(axis_1,axis_2,axis_3);
    auto gitter_one = *gitter1;
    BOOST_CHECK(gitter_one.axes[0]->cdscr ==  discretization::NONE);
    BOOST_CHECK(gitter_one.axes[0]->dscr ==  discretization::HOMOGENEOUS);

    gitter_one.template get_coords<false>(ttmp);

//     auto pgitter1 = std::shared_ptr<decltype(gitter1)>(&gitter1);
    auto full_iterator = gitter1->begin<false,true>();
//     debug::msg3 << full_iterator.begin << full_iterator.end << "strides:" << std::endl;
    debug::msg3 << gitter_one.data_info() << std::endl;
    do{
        *(full_iterator.ptr) = full_iterator.crds[0] + full_iterator.crds[1]/10.;
//         std::clog << *(full_iterator.ptr) << std::endl;
    } while((++full_iterator).within());
    std::array<T,392> target {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1.1,2.1,3.1,4.1,0,0,0,1.2,2.2,3.2,4.2,0,0,0,1.3,2.3,3.3,4.3,0,0,0,1.4,2.4,3.4,4.4,0,0,0,1.5,2.5,3.5,4.5,0,0,0,1.6,2.6,3.6,4.6,0,0,0,0,0,0,0,0,0,0,1.1,2.1,3.1,4.1,0,0,0,1.2,2.2,3.2,4.2,0,0,0,1.3,2.3,3.3,4.3,0,0,0,1.4,2.4,3.4,4.4,0,0,0,1.5,2.5,3.5,4.5,0,0,0,1.6,2.6,3.6,4.6,0,0,0,0,0,0,0,0,0,0,1.1,2.1,3.1,4.1,0,0,0,1.2,2.2,3.2,4.2,0,0,0,1.3,2.3,3.3,4.3,0,0,0,1.4,2.4,3.4,4.4,0,0,0,1.5,2.5,3.5,4.5,0,0,0,1.6,2.6,3.6,4.6,0,0,0,0,0,0,0,0,0,0,1.1,2.1,3.1,4.1,0,0,0,1.2,2.2,3.2,4.2,0,0,0,1.3,2.3,3.3,4.3,0,0,0,1.4,2.4,3.4,4.4,0,0,0,1.5,2.5,3.5,4.5,0,0,0,1.6,2.6,3.6,4.6,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};

//     for(auto& el: gitter1->storage){ std::clog << "," << el;}; std::clog << std::endl; //print the given array
    BOOST_TEST(gitter1->storage == target, boost::test_tools::per_element() );
}

auto gitter1 = make_grid<T,I,false>(axis_1,axis_2,axis_3);
//  debug::msg3 << gitter1.data_info() << std::endl;
BOOST_AUTO_TEST_CASE( grid_boundaries ) {
    gitter1.function<false,false,true>([](const std::array<T,3>& ar){return ar[1]+ar[0]/10.+ar[2]/100.;});
    gitter1.function<false,false,true>(testfunction<T>);
    debug::msg3 << gitter1 << std::endl;

    BOOST_CHECK_EQUAL(test3.function({-0.25,-0.2,0}),1.13);
    BOOST_CHECK_EQUAL(gitter1.axes[0]->boundaries[0].function({-0.25,-0.2,0}),1.13);

    debug::msg3 << gitter1.print_all() << std::endl;
    debug::msg3 << gitter1.print({3,0,0},{4,7,8}) << std::endl; // cut along z
    debug::msg3 << "Updating boundaries now." << std::endl;
    gitter1.update_boundaries();
    debug::msg3 << gitter1.print_all() << std::endl;
    debug::msg3 << gitter1.print({3,0,0},{4,7,8}) << std::endl; // cut along z
    BOOST_CHECK_EQUAL(gitter1.storage[7], 1.13); // DIRICHLET @ {0,1,#}
    BOOST_CHECK_EQUAL(gitter1.storage[100], -0.1886666666666667); // neumann @ {2,0,2}
    BOOST_CHECK_EQUAL(gitter1.storage[111], 1.5066666666666666); // neumann @ {1,7,2}
    BOOST_CHECK_EQUAL(gitter1.storage[304], 0.066666666666666666); // periodic @ {3,1,6}
}

BOOST_AUTO_TEST_CASE( exporting ){
    auto output = EXPORT::netcdf(gitter1, "test.nc");
    output.write_grid(gitter1);
}

BOOST_AUTO_TEST_CASE( sub_grids ){
    // base grid in "x,y,z"
    auto gitter1 = make_grid<T,I>(axis_1,axis_2,axis_3);
    debug::msg3 << discr_name(gitter1.axes[0]->dscr) << std::endl;

    auto tsst = gitter1.axes[0]->template shrink_stupid<std::vector<T>>();
    auto sub_props = gitter1.props.template sub_space<dim<'x','z'>>();
    auto sub_gitter1 = gitter1.template sub_grid<dim<'y','z'>>();
    debug::msg3 << "Full space: "<< gitter1 << std::endl;
    debug::msg3 << "subspace: "<< sub_gitter1 << std::endl;
}

BOOST_AUTO_TEST_CASE( reductions ){
    auto gitter1 = make_grid<T,I>(axis_1,axis_2,axis_3);
    gitter1.template function<false,false,false>([]()->T{return 1.;});

    auto reduced = gitter1.template reduction<reduction::SUM,dim<'x','y'>,false,true>([](T v, const std::array<T,3>&)->T{ return v;});
//     auto reduced = gitter1.template reduction<reduction::SUM, dim<'x','y'>, false, true, std::function<T(T, const std::array<T,3>&)>>([](T v, const std::array<T,3>& c)->T{ return v;});

    debug::msg3 << gitter1.print_all() << std::endl << reduced.print_all() << std::endl;

    boundary::boundary<T,I,std::array<T,2>> new_bc(boundary::btype::NEUMANN,1,3.33);
    reduced.change_boundaries('x',false,new_bc);
    reduced.change_boundaries('y',false,new_bc);
    reduced.change_boundaries('x',true,new_bc);
    debug::msg3 << reduced.print_all() << std::endl;
    reduced.update_boundaries();
    debug::msg3 << reduced.print_all() << std::endl;
}

BOOST_AUTO_TEST_CASE( reduction_with_direct_iter ){
    debug::msg2 << "here" <<std::endl;
    auto gitter1 = make_grid<T,I>(axis_1,axis_2,axis_3);
    gitter1.template function<false,false,false>([]()->T{return 1.;});

    auto reduced = gitter1.template reduction<reduction::SUM,dim<'x','y'>,iter_type<false,true>::template delta_coords>([](const auto& c)->T{ return *c +c.coords[0];});
    auto reduced2 = gitter1.template reduction<reduction::SUM,dim<'x','y'>,basic_iterator>([](const auto& c)->T{ return *c;});
    //TODO: eventhough basic_iterator is the default template parameter, this does not work without a specific value! WHY?

    debug::msg2 << gitter1.print_all() << std::endl << reduced.print_all() << std::endl;

    boundary::boundary<T,I,std::array<T,2>> new_bc(boundary::btype::NEUMANN,1,3.33);
    reduced.change_boundaries('x',false,new_bc);
    reduced.change_boundaries('y',false,new_bc);
    reduced.change_boundaries('x',true,new_bc);
    debug::msg2 << reduced.print_all() << std::endl;
    reduced.update_boundaries();
    debug::msg2 << reduced.print_all() << std::endl;
}

BOOST_AUTO_TEST_SUITE_END()
