#pragma once

#include <array>
#include <vector>

#include <sstream>

#include "optimize.h"

template<typename T, typename I>
struct csr{
    csr(size_t rows=0, size_t NNZ=0){
        IA.reserve(rows+1);
        IA[0]=0;
        A.reserve(NNZ);
    }

    std::vector<size_t> IA = {0}; // += number of non-zero entries per row
    std::vector<I> JA; // column index
    std::vector<T> A; // non zero entries

    T operator()(const size_t& n, const size_t& m) const {
        auto finder = std::find(JA.cbegin() + IA[n], JA.cbegin()+IA[n+1], m);
        if (finder == JA.cbegin()+IA[n+1]){
            return 0;
        }
        return *(A.cbegin()+(finder-JA.cbegin()));
    }

    csr<T,I>& operator+(const csr<T,I>& _in);
    csr<T,I>& operator*(T factor);
    csr<T,I>& operator/(T factor);

    size_t append_row(const std::vector<I>& _pos, const std::vector<T>& _val);
    size_t append_empty_row();

    std::string to_dense() const;
    template<char SEPARATOR = ','> // python: , ; matlab: ;
    std::string to_python_matrix() const;

    void shrink_to_fit(){
        IA.shrink_to_fit();
        JA.shrink_to_fit();
        A.shrink_to_fit();
    }

    void min_max_column(I& min, I& max) const;
protected:
};


template <typename T, typename I>
std::ostream& operator<<(std::ostream& o, const csr<T,I>& arr)
{
    o <<std::endl << arr.to_dense();
    return o;
}

#include "csr.hpp"
