#pragma once

#include "csr.h"

template<typename T, typename I>
csr<T,I>& csr<T,I>::operator*(T factor){
    auto output = *this;
    for(auto& el: output.A){
        el*=factor;
    }
    return output;
}

template<typename T, typename I>
csr<T,I>& csr<T,I>::operator/(T factor){
    assert(!is_about_zero(factor));

    auto output = *this;
    for(auto& el: output.A){
        el/=factor;
    }
    return output;
}

template<typename T, typename I>
csr<T,I>& csr<T,I>::operator+(const csr<T,I>& _in){
    assert(_in.IA.size() == IA.size());

    csr<T,I> output;
    output.IA.resize(IA.size());
    IA[0]=0;

    size_t s1=0, s2=0;
    size_t IA_o=0;
    auto i1 = IA.cbegin()+1;
    auto i2 = _in.IA.cbegin()+1;
    auto io = output.IA.begin()+1;
    for(; i1!=IA.cend(); ++i1, ++i2, ++io){
        for(size_t j1=s1; j1<*i1; ++j1){
            for(size_t j2=s2; j2<*i2; ++j2){
                if(JA[j1]==_in.JA[j2]){
                    output.JA.push_back(JA[j1]);
                    output.A.push_back(A[j1]+_in.A[j2]);
                    goto csr_e1;
                }
            }
            output.JA.push_back(JA[j1]);
            output.A.push_back(A[j1]);
            csr_e1:
            IA_o++;
        }

        *io=IA_o;
        s1=*i1;
        s2=*i2;
    }

    return output;
}

template<typename T, typename I>
size_t csr<T,I>::append_row(const std::vector<I>& _pos, const std::vector<T>& _val){
    assert(_pos.size() == _val.size());
//     debug::msg3 << A << "+" << _val << " -> " ;
    JA.insert(JA.end(), _pos.begin(), _pos.end());
    A.insert(A.end(), _val.begin(), _val.end());
//     debug::msg3 << A << std::endl;
    IA.push_back(IA.back()+_pos.size());
    return IA.back();
}

template<typename T, typename I>
void csr<T,I>::min_max_column(I& min, I& max) const {
    min=JA[0];
    max=JA[0];
    for(auto iter = JA.cbegin(); iter!=JA.cend(); ++iter){
        if(*iter < min){
            min = *iter;
        } else if (*iter > max) {
            max = *iter;
        }
    }
}

template<typename T, typename I>
std::string csr<T,I>::to_dense() const{
    std::ostringstream output;
    std::system("tabs 8");
    std::string separator = "\t";

    I pos, min, max;
    min_max_column(min,max);
    max++;

    size_t tmp = IA[0];
    for(auto ia_iter = IA.cbegin()+1; ia_iter!=IA.cend(); ++ia_iter){
        pos = min;
        for(; tmp<*ia_iter; ++tmp){
            for(; pos<JA[tmp]; ++pos){
                output << "0" << separator;
            }
            output << A[tmp] << separator;
            pos++;
        }
        for(; pos<max; ++pos){
            output << "0" << separator;
        }

        output << std::endl;
    }

    return output.str();
}

template<typename T, typename I>
template<char SEPARATOR>
std::string csr<T,I>::to_python_matrix() const{
    std::ostringstream output;
//     std::system("tabs 8");
    std::string separator = ",";

    I pos, min, max;
    min_max_column(min,max);
    max++;

    output << "[";
    size_t tmp = IA[0];
    for(auto ia_iter = IA.cbegin()+1; ia_iter!=IA.cend(); ++ia_iter){
        pos = min;
        output << "[";
        for(; tmp<*ia_iter; ++tmp){
            for(; pos<JA[tmp]; ++pos){
                output << "0" << separator;
            }
            output << A[tmp] << separator;
            pos++;
        }
        for(; pos<max; ++pos){
            output << "0" << separator;
        }

        output << "]";
        if (ia_iter!=IA.cend()-1){
            output << std::to_string(SEPARATOR);
        }
    }
    output << "]";

    return output.str();
}

template<typename T, typename I>
size_t csr<T,I>::append_empty_row(){
    IA.push_back(IA.back());
    return IA.back();
}
