#define BOOST_TEST_MODULE test_stencil
#include <boost/test/unit_test.hpp>

#include "debug.h"
#include "grid.h"
#include "axis.h"
#include "stencil.h"
#include "stencil_func.h"
#include "gmres.h"
#include "grid_iterator_delta_coords.h"

using T = double;
using I = long;

T tolerance = 1e5*std::numeric_limits<T>::epsilon();

BOOST_AUTO_TEST_SUITE( gmres_test , * boost::unit_test::tolerance(tolerance))


stencil<T,I,dim<'x'>,true> stx(1.,{1,-2,1},{{-1},{0},{1}});
stencil<T,I,dim<'y'>,true> sty(1.,{1,-2,1},{{-1},{0},{1}});
auto laplace2d = stx.kron(sty.unit_stencil()) + stx.unit_stencil().kron(sty);

boundary::boundary<T,I,std::array<T,2>> mirror_bc(boundary::btype::PERIODIC,1);
boundary::boundary<T,I,std::array<T,2>> dirichlet_bc(boundary::btype::DIRICHLET,1,0.);
auto x_axis = make_axis<T,I,'x'>(-.5,.5,12,dirichlet_bc,dirichlet_bc);
auto y_axis = make_axis<T,I,'y'>(-.5,.5,12,dirichlet_bc,dirichlet_bc);
auto square_grid = make_grid<T,I,false>(x_axis,y_axis);
auto lplc2d = (laplace2d.unit_stencil() - 0.5*laplace2d).adjust_to_grid(square_grid);

BOOST_AUTO_TEST_CASE( gmres_test ){
//     square_grid.template function<basic_iterator>([](const auto& c){ if(c.crds[0]==c.crds[1]){ return 1.; } else { return 0.;} });
    square_grid.template function<iter_type<false,true>::template delta_coords>([](const auto& c){ if(c.coords[0]*c.coords[0]+c.coords[1]*c.coords[1]<0.2){ return 1.; } else { return 0.;} });

    auto intermediate = lplc2d*square_grid;
    auto output(intermediate);

//     debug::msg2 << lplc2d.to_csr(square_grid) << std::endl;
//     debug::msg2 << lplc2d.to_csr(square_grid).to_python_matrix() << std::endl;
//     debug::msg2 << square_grid.storage << std::endl;

    debug::msg2 << "Iterations: " << gmres(lplc2d,intermediate,1000,20)(output) << std::endl;
    output = gmres(lplc2d,intermediate,1000,20)();

    debug::msg2 << "start grid:\n" << square_grid.print_all() << std::endl;
    debug::msg2 << "changed grid:\n" << intermediate.print_all() << std::endl;
    debug::msg2 << "gmres grid output:\n" << output.print_all() << std::endl;
//     debug::msg2 << lplc2d << std::endl;
//     debug::msg2 << lplc2d.to_csr(square_grid) << std::endl;

    BOOST_CHECK(intermediate.storage != square_grid.storage);
    BOOST_TEST(output.storage == square_grid.storage, boost::test_tools::per_element());
}

BOOST_AUTO_TEST_SUITE_END()
