#pragma once

#include <limits>

#include "stencil_func.h"
#include "grid.h"
#include "norms.h"

template<typename T, typename I, typename GRID, typename STENCIL, template <typename> class NORM = L2>
size_t richardson(GRID& output,
                  STENCIL& _stencil,
                  const GRID& _b,
                  T precision = 1e-5,
                  const size_t max_iterations = 1000,
                  const T relaxation = .1
                 ){
    size_t counter = 0;

    T error = 1000*precision;
    T nrm = 0.;

    GRID tmp = output;
    auto iter_b = _b.storage.cbegin();
    auto iter_t = tmp.storage.cbegin();
    auto iter_o = output.storage.begin();
    T val=0.;
    for(;counter < max_iterations; ++counter){
        tmp = _stencil*output;
        iter_b = _b.storage.cbegin();
        iter_t = tmp.storage.cbegin();
        iter_o = output.storage.begin();
        error=0.;
        nrm=0.;
        for(;iter_b!=_b.storage.cend(); ++iter_b, ++iter_t, ++iter_o){
            val = relaxation*(*iter_b - *iter_t);
// debug::msg2 << val << std::endl;
            error+=norm<T,NORM>::direct(val);
            *iter_o+=val;
            nrm+=norm<T,NORM>::direct(*iter_o);
        }
        val/=_b.storage.size();
        error=norm<T,NORM>::back(error);
        nrm=norm<T,NORM>::back(nrm);
// debug::msg2 << error/nrm << "/" << precision << std::endl;
        if (error<1*std::numeric_limits<T>::epsilon() || error/nrm < precision) {
            break;
        };
    }

//     assert(counter < max_iterations && "Richardson reached the maximum number of iterations.");
//     output.update_boundaries();
    return counter;
}
