#define BOOST_TEST_MODULE test_stencil
#include <boost/test/unit_test.hpp>

#include <cmath>
#include <iostream>
#include <string>
#include <limits>

#include "debug.h"
#include "grid.h"
#include "axis.h"
#include "stencil.h"
#include "richardson.h"

using T = double;
using I = long;


T tolerance = 10*std::numeric_limits<T>::epsilon();

BOOST_AUTO_TEST_SUITE( methods , * boost::unit_test::tolerance(tolerance))

stencil<T,I,dim<'x'>,true> stx(1.,{1,-2,1},{{-1},{0},{1}});
stencil<T,I,dim<'y'>,true> sty(1.,{1,-2,1},{{-1},{0},{1}});
auto laplace2d = stx.kron(sty.unit_stencil()) + stx.unit_stencil().kron(sty);

boundary::boundary<T,I,std::array<T,2>> mirror_bc(boundary::btype::PERIODIC,1);
boundary::boundary<T,I,std::array<T,2>> dirichlet_bc(boundary::btype::DIRICHLET,1,0.);
auto x_axis = make_axis<T,I,'x'>(-.5,.5,10,dirichlet_bc,dirichlet_bc);
auto y_axis = make_axis<T,I,'y'>(-.5,.5,10,dirichlet_bc,dirichlet_bc);
auto square_grid = make_grid<T,I,false>(x_axis,y_axis);
auto lplc2d = (laplace2d.unit_stencil() - 0.5*laplace2d).adjust_to_grid(square_grid);

BOOST_AUTO_TEST_CASE( richardson_test ){
    square_grid.function<false,false,true>([](const std::array<T,2>& ar){ if(is_about_zero(std::abs(ar[0]-ar[1]))){ return 1.; } else { return 0.;} });
//     square_grid.update_boundaries();

//     square_grid.function<false,false,true>([](const std::array<T,2>& ar){ return ar[0]+ar[1]/10.; });

    auto intermediate = lplc2d*square_grid;
    auto output(intermediate);



//     lplc2d.direct_inv(intermediate);
//     auto output2 = lplc2d.direct_inv(intermediate);
//     BOOST_TEST(output2.storage == square_grid.storage, boost::test_tools::per_element() );

    debug::msg3 << square_grid.print_all() << std::endl;
    debug::msg3 << intermediate.print_all() << std::endl;
    debug::msg3 << "Iterations: " << richardson<T,I>(output, lplc2d, intermediate, tolerance/10,1e6) << std::endl;
    debug::msg3 << output.print_all() << std::endl;
    BOOST_CHECK(intermediate.storage != square_grid.storage);
    BOOST_TEST(output.storage == square_grid.storage, boost::test_tools::per_element() );
//     debug::msg2 << lplc2d.to_csr(square_grid) << std::endl;
}

BOOST_AUTO_TEST_SUITE_END()
