#define BOOST_TEST_MODULE test_stencil
#include <boost/test/unit_test.hpp>

#include <cmath>
#include <iostream>
#include <string>
#include <limits>

#include "debug.h"
#include "grid.h"
#include "axis.h"
#include "stencil.h"

using T = double;
using I = long;


T tolerance = 10*std::numeric_limits<T>::epsilon();

BOOST_AUTO_TEST_SUITE( test_matrix , * boost::unit_test::tolerance(tolerance))

using dm = dim<'x','y','z'>;
using st = stencil<T,I,dm,true>;


T dirichlet1 = 0.33;
T dirichlet2 = 1.57;
T neumann1 = 2.57;

//\TODO: The boundary functions currently have to be of the same type. This should be more general, i.e. the most general, suitable type should be used.
boundary::boundary<T,I,std::array<T,2>> boundary_end(boundary::btype::OPEN,0);
boundary::boundary<T,I,std::array<T,2>> boundary_begin(boundary::btype::DIRICHLET,1,[](const std::array<T,2>&){return 1;});
boundary::boundary<T,I,std::array<T,2>> boundary_neumann(boundary::btype::NEUMANN,2,[](const std::array<T,2>&){return 0.;});

auto axis_z = make_axis<T,I,'z'>(0., 1.,100,boundary_begin,boundary_end);
auto axis_mu = make_axis<T,I,'m'>(0., 1.,101,boundary_end,boundary_neumann);
auto grid = make_grid<T,I>(axis_z, axis_mu);


T xi = 1.;
T kappa = .27;
T lambda = .027;

auto adv_z_tmp = stencil<T,I,dim<'z','m'>,true>(xi,{-1.,1.},{{-1,0},{0,0}});
auto adv_z = adv_z_tmp.get_delta_coords_f([](T val, const std::array<T,2>& delta, const std::array<T,2>& coords)->T {return val*coords[1]/delta[0];});
auto adv_mu_tmp = stencil<T,I,dim<'z','m'>,true>(kappa/xi,{1.,-2.,1.},{{0,-1},{0,0},{0,1}});
auto adv_mu = adv_mu_tmp.get_delta_coords_f(([](T val, const std::array<T,2>& delta, const std::array<T,2>& coords)->T {return val*(1-coords[0]*coords[0])/coords[1]/(delta[1]*delta[1]);}));

auto adv = adv_z-adv_mu;

auto diff_mu_l = stencil<T,I,dim<'z','m'>,true>(1., {1.}, {{0,-1}}).get_delta_coords_f([](T val, const std::array<T,2>& delta, const std::array<T,2>& coords)->T {return val*(1-std::pow(coords[1]-delta[1]/2.,2))/delta[1]/delta[1];});
auto diff_mu_u = stencil<T,I,dim<'z','m'>,true>(1., {1.}, {{0,1}}).get_delta_coords_f([](T val, const std::array<T,2>& delta, const std::array<T,2>& coords)->T {return val*(1-std::pow(coords[1]+delta[1]/2.,2))/delta[1]/delta[1];});
auto diff_mu_c = stencil<T,I,dim<'z','m'>,true>(2., {1.}, {{0,0}}).get_delta_coords_f([](T val, const std::array<T,2>& delta, const std::array<T,2>& coords)->T {return val*(1-std::pow(coords[1],2)-std::pow(delta[1],1))/delta[1]/delta[1];});
auto diff_mu = diff_mu_l+diff_mu_c+diff_mu_u;

auto A = adv-diff_mu;

BOOST_AUTO_TEST_CASE( basic ){ }


BOOST_AUTO_TEST_SUITE_END()
