#ifndef dim_h
#define dim_h

#include <array>
#include "liste.h"

template<char ... names>
using dim = liste<char, names...>;


template<typename C, C _search, C ... _liste>
struct find_in_pack;

template<typename C, C _f, C f0, C ... _fs>
struct find_in_pack<C, _f, f0, _fs...>{
    const static int pos = 1 + find_in_pack<C,_f,_fs...>::pos;
};

template<typename C, C _f, C ... _fs>
struct find_in_pack<C, _f, _f, _fs...>{
    const static int pos = 0;
};

template<size_t pos, class lower_limits, class upper_limits, class sizes>
struct boundary_limits{
    using lower_begin = typename liste_<size_t, lower_limits::size,0>::fill;
    using lower_end = typename sizes::template mix_at<pos,lower_limits>;
    using upper_begin = typename lower_begin::template mix_at<pos,upper_limits>;
    using upper_end = sizes;
};

template<bool HD, class lower_gc, class points, class upper_gc>
struct grid_props{
    constexpr static size_t dim = points::size;
    constexpr static bool hd = HD;
    using limit_lower = lower_gc;
    using lGC = lower_gc;
    using uGC = upper_gc;
    using pts = points;
    using limit_upper = typename lists_add<size_t,lower_gc,points>::reslt;
    using sizes = typename lists_add<size_t,limit_upper,upper_gc>::reslt;
    using strides = typename liste2strides<sizes>::template list<1>;
    constexpr static size_t size = liste2strides<sizes>::size;

    template <bool HD_new, size_t lgc, size_t pnts, size_t ugc>
    using append=grid_props<HD&&HD_new,
                 typename lower_gc::template append<lgc>,
                 typename points::template append<pnts>,
                 typename upper_gc::template append<ugc>>;
};
template<int size, bool HD>
using zero_grid_props = grid_props<HD, typename liste_<size_t,size,0>::fill, typename liste_<size_t,size,1>::fill, typename liste_<size_t,size,0>::fill>;

template<int pos, typename T, T val, T...targs>
struct nth_element{
    const static T value = nth_element<pos-1,T,targs...>::value;
    const static T product = val * nth_element<pos-1,T,targs...>::product;
};
template<typename T, T val, T...targs>
struct nth_element<0,T,val,targs...>{
    const static T value = val;
    const static T product = val;
};

template<typename V=int, class ... axes>
struct axes2type;

template<typename V, class axis, class ... axes>
struct axes2type<V, axis, axes...>{
    template<char ... _names>
    using names = typename axes2type<V, axes...>::template names<_names...,axis::name>;

    template<V last=1, V ... _strides>
    using strides = typename axes2type<V, axes...>::template strides<last*axis::size, _strides..., last*axis::size>;

    constexpr static bool homogeneous = axis::homogeneous && axes2type<V, axes...>::homogeneous;

    constexpr static int size = axes2type<V,axes...>::size;
};
template<typename V>
struct axes2type<V>{
    template<char ... _names>
    using names = liste<char,_names...>;
    template<V last, V ... _strides>
    using strides =liste<V,1,_strides...>;

    constexpr static bool homogeneous = true;
    constexpr static int size = 0;
};


#endif
