#include <iostream>
#include <vector>

#include <Eigen/Dense>
#include <Eigen/StdVector>
#include <Eigen/Core>
#include <chrono>

#include "grid.h"
#include "stencil.h"

using T = double;
using I = long;


using Eigen::MatrixXd;
int main()
{
  MatrixXd m(2,2);
  m(0,0) = 3;
  m(1,0) = 2.5;
  m(0,1) = -1;
  m(1,1) = m(1,0) + m(0,1);
  std::cout << m.inverse() << std::endl;

  std::vector<double> out = {-2.,1.,0.,0.,
      1.,-2.,1.,1.,
      0.,1.,-2.,1.,
      0.,0.,1.,-2.};

  std::vector<double> vec = {1.,1.,1.,1.,2.,2.,2.,2.};

//       Eigen::Map<Eigen::MatrixXi> matrix(out.data(),4,4,Eigen::OuterStride<>(4));
      MatrixXd matrix = Eigen::Map<Eigen::Matrix<double,Eigen::Dynamic,Eigen::Dynamic,Eigen::RowMajor>>(out.data(),4,4);
    std::cout << matrix << std::endl;
    std::cout << matrix.inverse() << std::endl;

    Eigen::Map<Eigen::VectorXd > vector(vec.data(),4);

    std::cout << matrix.inverse() * vector << std::endl;


    //test
    boundary::boundary<T,I,std::array<T,3>> test4(boundary::btype::MIRROR,1);
    auto axis_x = make_axis<T,I,'x'>(0., 1.,100,test4,test4);
    auto axis_y = make_axis<T,I,'y'>(0., 1.,100,test4,test4);
    auto axis_z = make_axis<T,I,'z'>(0., 1.,100,test4,test4);

    auto grid1 = make_grid<T,I>(axis_x,axis_y,axis_z);

    stencil<T,I,dim<'x'>> stx(1.,{1,-2,1},{{-1},{0},{1}});
    stencil<T,I,dim<'y'>> sty(1.,{1,-2,1},{{-1},{0},{1}});
    stencil<T,I,dim<'z'>> stz(1.,{1,-2,1},{{-1},{0},{1}});

    auto st_p = (stencil<T,I,dim<'x','y','z'>>(1.,{1,1,1,1,1,1,-8},{{-1,0,0},{1,0,0},{0,-1,0},{0,1,0},{0,0,-1},{0,0,1},{0,0,0}}));
    auto st = st_p.adjust_to_grid(grid1);

    debug::msg2 << st << std::endl;
    auto t_begin = std::chrono::high_resolution_clock::now();
    for (int i=0; i<100; ++i){
        st*grid1;
    }
    auto duration= std::chrono::duration_cast<std::chrono::milliseconds>(std::chrono::high_resolution_clock::now() - t_begin);
    std::cout << duration.count() << "ms"<< std::endl;

//     auto csr = st.to_csr(grid1);
    return 0;
}
