#pragma once

#include <type_traits>
#include <iostream>
#include <tuple>
#include <utility>

#include "helpers.h"
#include "debug.h"

template<class output, class input>
std::vector<output> transform_vector(const std::vector<input>& _input){
    if constexpr (std::is_same<input,output>::value) { return _input;}

    std::vector<output> _output(_input.size());

    auto iter_in = _input.cbegin();
    auto iter_out = _output.begin();

    for(; iter_in != _input.cend(); ++iter_in, ++iter_out){
        *iter_out = output(*iter_in);
    }

    return _output;
}

template<class names, typename T, typename I, typename ... secondary>
std::function<T(T, const secondary& ...)> functional_deflt = [](T input, const secondary& ...) {return input;};
template<class names, typename T, typename I, typename ... secondary>
inline T function_deflt(T input, const secondary& ...) {return input;}
std::string default_function_descriptor = "default passthrough";


/////////  deducing the stencil_func template parameters from paramter switches

template<class names, typename T, typename I, bool is_function, bool is_deltas, bool is_coords, bool HD, template <typename,typename,class,bool,bool,bool,bool, class ... ffargs> class f_class, class ... f_args>
struct decide_stencil_func_type;
template<class names, typename T, typename I, bool is_function, bool HD, template <typename,typename,class,bool,bool,bool,bool, class ... ffargs> class f_class>
struct decide_stencil_func_type<names,T,I,is_function,false,false,HD,f_class>{
    using type = f_class<T,I,names,HD,is_function,false,false>;
};
template<class names, typename T, typename I, bool is_function, template <typename,typename,class,bool,bool,bool,bool, class ... ffargs> class f_class>
struct decide_stencil_func_type<names,T,I,is_function,true,false,false,f_class>{
    using type = f_class<T,I,names,false,is_function,true,false,std::array<T,names::size>>;
};
template<class names, typename T, typename I, bool is_function, template <typename,typename,class,bool,bool,bool,bool, class ... ffargs> class f_class>
struct decide_stencil_func_type<names,T,I,is_function,true,false,true,f_class>{
    using type = f_class<T,I,names,true,is_function,false,false>;
};
template<class names, typename T, typename I, bool is_function, bool HD, template <typename,typename,class,bool,bool,bool,bool, class ... ffargs> class f_class>
struct decide_stencil_func_type<names,T,I,is_function,false,true,HD,f_class>{
    using type = f_class<T,I,names,HD,is_function,false,true,std::array<T,names::size>>;
};
template<class names, typename T, typename I, bool is_function, template <typename,typename,class,bool,bool,bool,bool, class ... ffargs> class f_class>
struct decide_stencil_func_type<names,T,I,is_function,true,true,false,f_class>{
    using type = f_class<T,I,names,false,is_function,true,true,std::array<T,names::size>,std::array<T,names::size>>;
};
template<class names, typename T, typename I, bool is_function, template <typename,typename,class,bool,bool,bool,bool, class ... ffargs> class f_class>
struct decide_stencil_func_type<names,T,I,is_function,true,true,true,f_class>{
    using type = f_class<T,I,names,true,is_function,false,true,std::array<T,names::size>>;
};


// determining the function signature from the template parameters
// TODO: merge with the above decide_stencil_func_type
template<size_t dim, typename T, typename I, typename VALUE, bool access, bool is_deltas, bool is_coords>
struct function_type;
template<size_t dim, typename T, typename I, typename VALUE>
struct function_type<dim,T,I,VALUE,true,true,true>{ using type = std::function<T(VALUE,const std::array<T,dim>&,const std::array<T,dim>&)>; };
template<size_t dim, typename T, typename I, typename VALUE>
struct function_type<dim,T,I,VALUE,true,true,false>{ using type = std::function<T(VALUE,const std::array<T,dim>&)>;};
template<size_t dim, typename T, typename I, typename VALUE>
struct function_type<dim,T,I,VALUE,true,false,true>{ using type = std::function<T(VALUE,const std::array<T,dim>&)>;};
template<size_t dim, typename T, typename I, typename VALUE>
struct function_type<dim,T,I,VALUE,true,false,false>{ using type = std::function<T(VALUE)>;};
template<size_t dim, typename T, typename I, typename VALUE>
struct function_type<dim,T,I,VALUE,false,true,true>{ using type = std::function<T(const std::array<T,dim>&,const std::array<T,dim>&)>; };
template<size_t dim, typename T, typename I, typename VALUE>
struct function_type<dim,T,I,VALUE,false,true,false>{ using type = std::function<T(const std::array<T,dim>&)>;};
template<size_t dim, typename T, typename I, typename VALUE>
struct function_type<dim,T,I,VALUE,false,false,true>{ using type = std::function<T(const std::array<T,dim>&)>;};
template<size_t dim, typename T, typename I, typename VALUE>
struct function_type<dim,T,I,VALUE,false,false,false>{ using type = std::function<T()>;};

template<size_t dim, typename T, typename I, typename VALUE, bool access, class ITER_TYPE>
struct function_type_e{
    using type = typename  function_type<dim,T,I,VALUE,access,ITER_TYPE::uses_deltas,ITER_TYPE::uses_coords>::type;
};


// transform the functions

template<class names, typename T, typename I, bool is_deltas_out, bool is_coords_out, bool is_deltas_in, bool is_coords_in, bool HD, class ... f_args>
struct transform_function;

// TODO: get the passthrough cases to use this code instead of the many specializations below
template<class names, typename T, typename I, bool is_deltas_in, bool is_coords_in, bool HD, class ... f_args>
struct transform_function<names,T,I,is_deltas_in, is_coords_in, is_deltas_in, is_coords_in, HD, f_args ...>{
    static constexpr std::function<T(T,const f_args&...)> trafo(const std::function<T(T, const f_args&...)>& _in, const std::array<T,names::size>&){ return _in;}
};

template<class names, typename T, typename I, bool HD>
struct transform_function<names,T,I,true, true, false, false, HD>{
    static constexpr std::function<T(T, const std::array<T,names::size>&, const std::array<T,names::size>&)> trafo(const std::function<T(T)>& _in, const std::array<T,names::size>&){ return [_in](T _tin, const std::array<T,names::size>&, const std::array<T,names::size>&){ return _in(_tin);};}
};
template<class names, typename T, typename I>
struct transform_function<names,T,I,true, true, true, false, false>{
    static constexpr std::function<T(T, const std::array<T,names::size>&, const std::array<T,names::size>&)> trafo(const std::function<T(T, const std::array<T,names::size>&)>& _in, const std::array<T,names::size>&){ return [_in](T _tin, const std::array<T,names::size>& _ard, const std::array<T,names::size>&){return _in(_tin,_ard);};}
};
template<class names, typename T, typename I>
struct transform_function<names,T,I,true, true, true, false, true>{
    static constexpr std::function<T(T, const std::array<T,names::size>&)> trafo(const std::function<T(T, const std::array<T,names::size>&)>& _in, const std::array<T,names::size>& _deltas){ return [_in,_deltas](T _tin, const std::array<T,names::size>&){return _in(_tin,_deltas);};}
};
template<class names, typename T, typename I, bool HD>
struct transform_function<names,T,I,true, true, false, true, HD>{
    static constexpr std::function<T(T, const std::array<T,names::size>&, const std::array<T,names::size>&)> trafo(const std::function<T(T, const std::array<T,names::size>&)>& _in, const std::array<T,names::size>&){ return [_in](T _tin, const std::array<T,names::size>&, const std::array<T,names::size>& _arc){return _in(_tin,_arc);};}
};
template<class names, typename T, typename I, bool HD>
struct transform_function<names,T,I,true, false, false, false, HD>{
    static constexpr std::function<T(T, const std::array<T,names::size>&)> trafo(const std::function<T(T)>& _in, const std::array<T,names::size>&){ return [_in](T _tin, const std::array<T,names::size>&){return _in(_tin);};}
};
template<class names, typename T, typename I>
struct transform_function<names,T,I, true, true, true, true, false>{
    static constexpr std::function<T(T, const std::array<T,names::size>&, const std::array<T,names::size>&)> trafo(const std::function<T(T, const std::array<T,names::size>&, const std::array<T,names::size>&)>& _in, const std::array<T,names::size>&){ return _in;}
};
template<class names, typename T, typename I>
struct transform_function<names,T,I, true, true, true, true, true>{
    static constexpr std::function<T(T, const std::array<T,names::size>&)> trafo(const std::function<T(T, const std::array<T,names::size>&, const std::array<T,names::size>&)>& _in, const std::array<T,names::size>& _deltas){ return [_in,_deltas](T _tin, const std::array<T,names::size>& _arc) { return _in(_tin, _deltas, _arc);};}
};
template<class names, typename T, typename I>
struct transform_function<names,T,I,true, false, true, false, false>{
    static constexpr std::function<T(T, const std::array<T,names::size>&)> trafo(const std::function<T(T, const std::array<T,names::size>&)>& _in, const std::array<T,names::size>&){ return [_in](T _tin, const std::array<T,names::size>& _ard){return _in(_tin,_ard);};}
};
template<class names, typename T, typename I>
struct transform_function<names,T,I,true, false, true, false, true>{
    static constexpr std::function<T(T)> trafo(const std::function<T(T, const std::array<T,names::size>&)>& _in, const std::array<T,names::size>& _deltas){ return [_in,_deltas](T _tin){return _in(_tin, _deltas);};}
};
template<class names, typename T, typename I, bool HD>
struct transform_function<names,T,I, false, true, false, false, HD>{
    static constexpr std::function<T(T, const std::array<T,names::size>&)> trafo(const std::function<T(T)>& _in, const std::array<T,names::size>&){ return [_in](T _tin, const std::array<T,names::size>&){ return _in(_tin);};}
};
template<class names, typename T, typename I, bool HD>
struct transform_function<names,T,I, false, true, false, true, HD>{
    static constexpr std::function<T(T, const std::array<T,names::size>&)> trafo(const std::function<T(T, const std::array<T,names::size>&)>& _in, const std::array<T,names::size>&){ return [_in](T _tin, const std::array<T,names::size>& _arc){return _in(_tin,_arc);};}
};
template<class names, typename T, typename I>
struct transform_function<names,T,I, false, true, true, false, true>{
    static constexpr std::function<T(T, const std::array<T,names::size>&)> trafo(const std::function<T(T, const std::array<T,names::size>&)>& _in, const std::array<T,names::size>& _delta){ return [_in,_delta](T _tin, const std::array<T,names::size>&){return _in(_tin,_delta);};}
};
template<class names, typename T, typename I>
struct transform_function<names,T,I, false, false, true, false, true>{
    static constexpr std::function<T(T)> trafo(const std::function<T(T, const std::array<T,names::size>&)>& _in, const std::array<T,names::size>& _delta){ return [_in,_delta](T _tin){return _in(_tin,_delta);};}
};
template<class names, typename T, typename I>
struct transform_function<names,T,I, false, true, true, true, true>{
    static constexpr std::function<T(T, const std::array<T,names::size>&)> trafo(const std::function<T(T, const std::array<T,names::size>&, const std::array<T,names::size>&)>& _in, const std::array<T,names::size>& _delta){ return [_in,_delta](T _tin, const std::array<T,names::size>& _arc){ return _in(_tin,_delta,_arc);};}
};

///// // size of a template pack
template<int ... pos>
struct count;
template<int e, int ... pos>
struct count <e,pos...>{
    const static std::size_t size = 1 + count<pos ...>::size;
};
template<>
struct count<> {
    const static std::size_t size = 0;
};

///////////////////////////

template<std::size_t N, std::size_t... Seq>
constexpr std::index_sequence<N + Seq ...>
add(std::index_sequence<Seq...>)
{ return {}; }

template<std::size_t Min, std::size_t Max>
using make_index_range = decltype(add<Min>(std::make_index_sequence<Max-Min>()));

///////////////////////////////

template<typename OAR, typename AR, size_t ... pos>
inline auto reducer (const AR& _in, std::index_sequence<pos...>){
//     size_t size = count<pos...>::size;
    return OAR(std::get<pos>(_in)...);
}
template<typename I, size_t indim, size_t begin, size_t length, typename Index = make_index_range<begin,begin+length> >
inline std::array<I, length> reduce ( const std::array<I, indim>& _in){
    return reducer<std::array<I, length>>(_in, Index {});
}


// expand_function
template<bool is_upper, int dim, int new_dim, typename T, typename I, class ... f_args>
struct expand_function;
template<bool is_upper, int dim, int new_dim, typename T, typename I>
struct expand_function<is_upper, dim, new_dim, T, I>{
    constexpr static inline std::function<T(T)> get(const std::function<T(T)>& _in) {return _in;};
};
template<int dim, int new_dim, typename T, typename I>
struct expand_function<true, dim, new_dim, T, I, std::array<T,dim>>{
    constexpr static inline std::function<T(T, const std::array<T,new_dim>&)> get(const std::function<T(T,const std::array<T,dim>&)>& _in) {
        return [=](T t, const std::array<T,new_dim>& ar) {return _in(t,reduce<T,new_dim,0,dim>(ar));};
    };
};
template<int dim, int new_dim, typename T, typename I>
struct expand_function<false, dim, new_dim, T, I, std::array<T,dim>>{
    constexpr static inline std::function<T(T,const std::array<T,new_dim>&)> get(const std::function<T(T,const std::array<T,dim>&)>& _in) {
        return [=](T t, const std::array<T,new_dim>& ar) {return _in(t,reduce<T,new_dim,new_dim-dim,dim>(ar));};
    };
};
template<int dim, int new_dim, typename T, typename I>
struct expand_function<true, dim, new_dim, T, I, std::array<T,dim>, std::array<T,dim>>{
    constexpr static inline std::function<T(T, const std::array<T,new_dim>&, const std::array<T,new_dim>&)> get(const std::function<T(T,const std::array<T,dim>&)>& _in) {
        return [=](T t, const std::array<T,new_dim>& ard, const std::array<T,new_dim>& arc) {return _in(t,reduce<T,new_dim,0,dim>(ard),reduce<T,new_dim,new_dim-dim,dim>(arc));};
    };
};
template<int dim, int new_dim, typename T, typename I>
struct expand_function<false, dim, new_dim, T, I, std::array<T,dim>, std::array<T,dim>>{
    constexpr static inline std::function<T(T, const std::array<T,new_dim>&, const std::array<T,new_dim>&)> get(const std::function<T(T,const std::array<T,dim>&)>& _in) {
        return [=](T t, const std::array<T,new_dim>& ard, const std::array<T,new_dim>& arc) {return _in(t,reduce<T,new_dim,new_dim-dim,dim>(ard),reduce<T,new_dim,new_dim-dim,dim>(arc));};
    };
};


// shift function
template<size_t dim, typename T, typename I, bool HD, bool is_deltas, bool is_coords>
struct shift_function;
template<size_t dim, typename T, typename I>
struct shift_function<dim,T,I,true,true,true>{
    constexpr static inline std::function<T(T, const std::array<T,dim>&, const std::array<T,dim>&)> by_offset(const std::function<T(T, const std::array<T,dim>&, const std::array<T,dim>&)>& _f, const std::array<T,dim>& shift){
        return [=](T v, const std::array<T,dim>& d, const std::array<T,dim>& c){return _f(v, d, c+shift);};
    }
};
template<size_t dim, typename T, typename I>
struct shift_function<dim,T,I,true,false,true>{
    constexpr static inline std::function<T(T, const std::array<T,dim>&)> by_offset(const std::function<T(T, const std::array<T,dim>&)>& _f, const std::array<T,dim>& shift){
        return [=](T v, const std::array<T,dim>& c){return _f(v, c+shift);};
    }
};


//reduce function
template<typename>
struct array_traits;
template<typename T, size_t S>
struct array_traits<std::array<T,S>>{
    using type = T;
    constexpr static size_t size = S;
};

template<size_t new_dim, typename T>
struct change_array_size{
    using type = std::array<typename array_traits<T>::type,new_dim>;
};

