#ifndef stencil_func_hpp
#define stencil_func_hpp

#include "helpers.h"
#include "stencil_func.h"
#include "stencil_func_helpers.h"
#include "functional_types.h"

// constructors

template<typename T, typename I, class names, bool HD, bool is_function, bool is_deltas, bool is_coords, class ... f_args>
stencil_func<T,I,names,HD,is_function,is_deltas,is_coords,f_args...>::stencil_func() :
             stencil_proto<T,I,names>(),
             deltas({}),
             stencil_coefficients(std::vector<T>(0)),
             stencils(std::vector<this_type>(0)),
             std_function(function_deflt<names,T,I,f_args...>), function_descriptor(default_function_descriptor) {}

template<typename T, typename I, class names, bool HD, bool is_function, bool is_deltas, bool is_coords, class ... f_args>
stencil_func<T,I,names,HD,is_function,is_deltas,is_coords,f_args...>::stencil_func(T coeffi, const std::array<I,names::size>& offset):
stencil_func<T,I,names,HD,is_function,is_deltas,is_coords,f_args...>(coeffi, std::vector<T>({1.}), std::vector<std::array<I,names::size>>({offset}) ) {}

template<typename T, typename I, class names, bool HD, bool is_function, bool is_deltas, bool is_coords, class ... f_args>
stencil_func<T,I,names,HD,is_function,is_deltas,is_coords,f_args...>::stencil_func(T coeffi, I offset):
stencil_func<T,I,names,HD,is_function,is_deltas,is_coords,f_args...>(coeffi, std::array<I,1>({{offset}}) ) {}

template<typename T, typename I, class names, bool HD, bool is_function, bool is_deltas, bool is_coords, class ... f_args>
stencil_func<T,I,names,HD,is_function,is_deltas,is_coords,f_args...>::stencil_func(
            T _factor, const std::vector<T>& _coefficients, const std::vector<std::array<I,dim>>& _positions,
            const std::array<T,dim>& _deltas, const std::vector<T>& _stencil_coefficients,
            const typename std::vector<this_type>& _stencils,
            const typename std::function<T(T, const f_args& ...)>& _function, const std::string& f_description) :
             stencil_proto<T,I,names>(_factor, _coefficients, _positions),
             deltas(_deltas),
             stencil_coefficients(_stencil_coefficients), stencils(_stencils), std_function(_function), function_descriptor(f_description) {
                 static_assert(szeof<f_args...>::value==(is_deltas+is_coords),"wrong number of arguments in f_args.");
            }

template<typename T, typename I, class names, bool HD, bool is_function, bool is_deltas, bool is_coords, class ... f_args>
stencil_func<T,I,names,HD,is_function,is_deltas,is_coords,f_args...>::stencil_func(
            const grid<T,I,names,HD>& _grid,
            T _factor, const std::vector<T>& _coefficients, const std::vector<std::array<I,dim>>& _positions,
            const std::vector<T>& _stencil_coefficients,
            const typename std::vector<this_type>& _stencils,
            const typename std::function<T(T, const f_args& ...)>& _function, const std::string& f_description) :
            stencil_proto<T,I,names>(_grid, _factor, _coefficients, _positions),
            deltas(_grid.deltas), stencil_coefficients(_stencil_coefficients), stencils(_stencils),
            std_function(_function), function_descriptor(f_description) {
                static_assert(szeof<f_args...>::value==(is_deltas+is_coords),"wrong number of arguments in f_args.");
            }

template<typename T, typename I, class names, bool HD, bool is_function, bool is_deltas, bool is_coords, class ... f_args>
template<bool is_function_in, bool is_deltas_in, bool is_coords_in, class ... f_args_in>
stencil_func<T,I,names,HD,is_function,is_deltas,is_coords,f_args...>::stencil_func(const stencil_func<T,I,names,HD,is_function_in,is_deltas_in,is_coords_in, f_args_in...>& _st) :
                    stencil_func<T,I,names,HD,is_function,is_deltas,is_coords,f_args...>(_st.factor, _st.coefficients, _st.positions, _st.deltas, _st.stencil_coefficients, transform_vector<this_type>(_st.stencils), transform_function<names,T,I,is_deltas,is_coords,is_deltas_in,is_coords_in,HD>::trafo(_st.std_function, _st.deltas) //NOTE: HD =false is used here to transform the function as is ... -> TODO: make more general by adding subtemplate HD // this work, because under normal circumstances, the problematic cases do not occur.
                    , _st.function_descriptor
            ) {
//                         debug::msg2 << "deltasinv0: " << _st.deltas << std::endl;
                        static_assert(szeof<f_args...>::value==(is_deltas+is_coords),"wrong number of arguments in f_args.");
            }

// function stuff

template<typename T, typename I, class names, bool HD, bool is_function, bool is_deltas, bool is_coords, class ... f_args>
template<typename ITER>
inline T stencil_func<T,I,names,HD,is_function,is_deltas,is_coords,f_args...>::fun_fun(ITER itr, const f_args& ... arguments) const {
    T output = 0;

    auto iter_coeffi = stencil_coefficients.cbegin();
    auto iter_stencils = stencils.cbegin();
    for(; iter_coeffi!= stencil_coefficients.cend(); ++iter_coeffi, ++ iter_stencils){
        output+= *iter_coeffi * iter_stencils->function(itr, arguments...);
    }

    return output;
}

template<typename T, typename I, class names, bool HD, bool is_function, bool is_deltas, bool is_coords, class ... f_args>
template<typename ITER>
inline T stencil_func<T,I,names,HD,is_function,is_deltas,is_coords,f_args...>::function(ITER _in, const f_args& ... arguments) const {
//     debug::msg3 << _in << std::endl;
    if constexpr (is_function){
        return std_function((stencil_proto<T,I,names>::function(_in) + this->factor * fun_fun<ITER>(_in, arguments ... )), arguments ...);
    } else {
        return stencil_proto<T,I,names>::template function<ITER>(_in);
    }
}


template<typename T, typename I, class names, bool HD, bool is_function, bool is_deltas, bool is_coords, class ... f_args>
template<bool uses_deltas, bool uses_coords, class ... f_args_in>
inline typename decide_stencil_func_type<names,T,I,true,uses_deltas||is_deltas,uses_coords || is_coords,HD,stencil_func>::type stencil_func<T,I,names,HD,is_function,is_deltas,is_coords,f_args...>::get_functional(const std::function<T(T,const f_args_in&...)>& _fin, const std::string& descriptor){
    debug::msg1 << "The stencil matrix functions should be linear in nature. The programmer has to check it!" << std::endl;
#warning "The stencil matrix functions should be linear in nature. The programmer has to check it!"
//     debug::msg2 << "deltasinv1: " << this->deltas << std::endl;
    using out_type = typename decide_stencil_func_type<names,T,I,true,uses_deltas||is_deltas,uses_coords || is_coords,HD,stencil_func>::type;
    return out_type(1., {}, {},
                    this->deltas, {1.}, {out_type(this->factor, this->coefficients, this->positions, this->deltas, stencil_coefficients, transform_vector<out_type>(stencils)
         , transform_function<names,T,I, uses_deltas||is_deltas, uses_coords || is_coords, uses_deltas, uses_coords, HD>::trafo(_fin, this->deltas), descriptor
    )});
}

template<typename T, typename I, class names, bool HD, bool is_function, bool is_deltas, bool is_coords, class ... f_args>
stencil_func<T,I,names,HD,is_function,is_deltas,is_coords,f_args...>& stencil_func<T,I,names,HD,is_function,is_deltas,is_coords,f_args...>::operator=(const this_type& _in) {
    if (this == &_in){
        return *this;
    }
    new (this) this_type(_in);
    return *this;
}

template<typename T, typename I, class names, bool HD, bool is_function, bool is_deltas, bool is_coords, class ... f_args>
template<bool HD_new>
typename decide_stencil_func_type<names,T,I,is_function,is_deltas&&!HD_new,is_coords,HD_new,stencil_func>::type  stencil_func<T,I,names,HD,is_function,is_deltas,is_coords,f_args...>::adjust_to_grid(const grid<T,I,names,HD_new>& _grid) const {
    // let's start slowly without adjusting the functions
//     using new_type = stencil_func<T,I,names,HD_new,is_function,is_deltas&&!HD_new,is_coords,f_args...>;
    using new_type = typename decide_stencil_func_type<names,T,I,is_function,is_deltas&&!HD_new,is_coords,HD_new,stencil_func>::type;

    std::vector<I> new_linear;
    new_linear.reserve(this->linear_positions.size());
    for(auto iter = this->positions.cbegin(); iter!=this->positions.cend(); ++iter) {
        new_linear.push_back(_grid.props.template inner_product<false>(*iter));
    }

    std::vector<T> new_coefficients = this->coefficients;
    std::vector<std::array<I,dim>> new_positions = this->positions;
    sort_v<dim,T,I>(new_linear,new_coefficients,new_positions);

    std::vector<new_type> stencils_new;
    stencils_new.reserve(stencils.size());
    for(auto iter_old = stencils.cbegin(); iter_old!= stencils.cend(); ++ iter_old){
        stencils_new.push_back(iter_old->template adjust_to_grid<HD_new>(_grid));
    }

//     debug::msg2 << "deltasinv2: " << _grid.deltas << std::endl;
    return new_type(_grid, this->factor, new_coefficients, new_positions, this->stencil_coefficients, stencils_new, transform_function<names,T,I,is_deltas&&!HD_new,is_coords,is_deltas,is_coords,HD_new>::trafo(this->std_function, _grid.deltas), this->function_descriptor );
}

template<typename T, typename I, class names, bool HD, bool is_function, bool is_deltas, bool is_coords, class ... f_args>
template<bool adjust_vars>
stencil_func<T,I,names,HD,is_function,is_deltas,is_coords,f_args...> stencil_func<T,I,names,HD,is_function,is_deltas,is_coords,f_args...>::shift(const std::vector<T>& coeffi, const std::vector<std::array<I,dim>>& _shift) const{

    // this stencil layers part
    std::vector<std::array<I,dim>> new_pos;
    std::vector<T> new_coeffi;
    auto i_c = coeffi.cbegin();
    auto i_p = _shift.cbegin();
    for(;i_c!=coeffi.cend(); ++i_c, ++i_p){
        add_uniq_stencil_points<T,I>(new_coeffi, new_pos, this->coefficients, vector_shift(this->positions,*i_p), *i_c);
    }

    std::function<T(T, const f_args& ...)> new_function;
    std::ostringstream new_descriptor;
    new_descriptor << function_descriptor;
    if constexpr (adjust_vars) {
        assert(_shift.size()==1 && "Currently the shift operator just supports variable adjustment for single shifts.");
        static_assert(HD, "Shifting a stencil only works with a homogeneous discretization. so it might be necessary to use adjust_to_grid on the stencil before.");
        if constexpr (is_coords){ // in this way only possible for homogeneous discretization
            auto offset = deltas;
            for(size_t i=0; i<dim; ++i){
                offset[i]*=_shift[0][i];
            }
            new_function = shift_function<dim,T,I,HD,is_deltas,is_coords>::by_offset(std_function,offset);
            new_descriptor << "_shifted(" << _shift[0] << ")";
        } else {
            new_function = std_function;
        }
    } else {
        new_function = std_function;
    }

    // the substencils
    std::vector<this_type> new_stencils;
    new_stencils.reserve(stencils.size());
    for(auto iter = stencils.cbegin(); iter!=stencils.cend(); ++iter){
        new_stencils.push_back(iter->template shift<adjust_vars>(coeffi, _shift));
    }

    return this_type(this->factor, new_coeffi, new_pos, deltas, stencil_coefficients, new_stencils, new_function, new_descriptor.str());
}

template<typename T, typename I, class names, bool HD, bool is_function, bool is_deltas, bool is_coords, class ... f_args>
template<bool adjust_vars>
stencil_func<T,I,names,HD,is_function,is_deltas,is_coords,f_args...> stencil_func<T,I,names,HD,is_function,is_deltas,is_coords,f_args...>::shift(const std::array<I,dim>& _shift) const {
    return shift(std::vector<T>{1.},std::vector<std::array<I,dim>>{_shift});
}


template<typename T, typename I, class names, bool HD, bool is_function, bool is_deltas, bool is_coords, class ... f_args>
template<bool linear_pos>
std::string stencil_func<T,I,names,HD,is_function,is_deltas,is_coords,f_args...>::print(std::string indent) const{
    std::ostringstream out;
    out.precision(2);
    std::string separator = "\t";
    if constexpr (is_function) {
        //NOTE: I have no idea how to efficiently get a correct representation of the function. The easiest way would be to assign an additional string descriptor
        out << "function(R";
        if (function_descriptor == default_function_descriptor){
            out << ")=R" << "\n";
        } else {
            array_element_head = "d";
            if constexpr (is_deltas && HD) out << ",[" << names::list << "]";
            if constexpr (is_deltas && !HD) out << "," << names::list;
            array_element_head = "";
            if constexpr (is_coords) out << "," << names::list;
            out << ")=" << function_descriptor << "\n";
        }
    }
    out << indent << std::scientific << this->factor << "*"  << std::fixed <<separator;

    for (const auto& name: names::list){
        out << name << separator;
    }
    if constexpr (linear_pos){
        out << "lin_pos" << separator;
    }
    out << "\n" << indent;

    out << "_SUM" << std::string((8*(dim+1+linear_pos)+2),'_');

    auto i_pos = this->positions.cbegin();
    auto i_coe = this->coefficients.cbegin();
    auto i_lp = this->linear_positions.cbegin();
    for(; i_pos != this->positions.cend(); ++i_pos, ++i_coe, ++i_lp){
        out << "\n";
        out << indent << std::scientific << *i_coe << std::fixed <<separator;

        for(auto& pos: *i_pos){
            out << pos << separator;
        }
        if constexpr (linear_pos){
            out << *i_lp << separator;
        }
    }

    auto iter_st_c = this->stencil_coefficients.cbegin();
    auto iter_st_p = this->stencils.cbegin();
    for(; iter_st_p!= this->stencils.cend(); ++iter_st_c, ++iter_st_p){
        out << "\n";
//         out << indent << "substencil:\n";
        out << "    " << *iter_st_c << "*" << iter_st_p->print(indent+"    ") << "\n";
    }

    return out.str();
}

template<typename T, typename I, class names, bool HD, bool is_function, bool is_deltas, bool is_coords, class ... f_args>
auto stencil_func<T,I,names,HD,is_function,is_deltas,is_coords,f_args...>::zero_stencil(){
    return stencil_func<T,I,names,HD,false,false,false>();
}

template<typename T, typename I, class names, bool HD, bool is_function, bool is_deltas, bool is_coords, class ... f_args>
auto stencil_func<T,I,names,HD,is_function,is_deltas,is_coords,f_args...>::unit_stencil(){
    return stencil_func<T,I,names,HD,false,false,false>(1.,liste_<I,names::size,0>::fill::list);
}

template<typename T, typename I, class names, bool HD, bool is_function, bool is_deltas, bool is_coords, class ... f_args>
void stencil_func<T,I,names,HD,is_function,is_deltas,is_coords,f_args...>::run_on_grid(grid<T,I,names,HD>& _in, grid<T,I,names,HD>& output) const {
    static_assert(szeof<f_args...>::value==(is_deltas+is_coords),"wrong number of arguments in f_args.");

//TODO: unify with grid::function
//NOTE: copy from grid

    auto range = _in.template begin_intern<is_deltas,is_coords>(_in.props.limit_lower,_in.props.limit_upper);
    do{
        if constexpr (is_deltas && is_coords) {
            output.storage[range.ptr-_in.storage.data()] = function<T*>(range.ptr, range.deltas, range.coords);
        } else if constexpr (is_deltas && !is_coords) {
            output.storage[range.ptr-_in.storage.data()] = function<T*>(range.ptr, range.deltas);
        } else if constexpr (!is_deltas && is_coords) {
            output.storage[range.ptr-_in.storage.data()] = function<T*>(range.ptr, range.coords);
        } else if constexpr (!is_deltas && !is_coords) {
            output.storage[range.ptr-_in.storage.data()] = function<T*>(range.ptr);
        }
//         debug::msg2 << * (range.ptr) << range.deltas << range.coords << "->" << output.storage[range.ptr-_in.storage.data()] << std::endl;
    } while((++range).within());
    output.age=_in.age+1; // just to secure object existence.
}

template<typename T, typename I, class names, bool HD, bool is_function, bool is_deltas, bool is_coords, class ... f_args>
void stencil_func<T,I,names,HD,is_function,is_deltas,is_coords,f_args...>::run_on_grid(grid<T,I,names,HD>& _grid) const {
    static_assert(szeof<f_args...>::value==(is_deltas+is_coords),"wrong number of arguments in f_args.");
//     static_assert(false, "This function has not been implemented yet.");
    assert(false && "Not yet implemented.");

//NOTE: copy from grid
    //TODO: to make this work, a ring cache must be implemented.

    auto range = _grid.template begin_intern<is_deltas,is_coords>(_grid.props.limit_lower,_grid.props.limit_upper);
    do{
        if constexpr (is_deltas && is_coords) {
            *(range.ptr) = function<T*>(range.ptr, range.deltas, range.coords);
        } else if constexpr (is_deltas && !is_coords) {
            *(range.ptr) = function<T*>(range.ptr, range.deltas);
        } else if constexpr (!is_deltas && is_coords) {
            *(range.ptr) = function<T*>(range.ptr, range.coords);
        } else if constexpr (!is_deltas && !is_coords) {
            *(range.ptr) = function<T*>(range.ptr);
        }
    } while((++range).within());
    _grid.age++; // just to secure object existence.
}


template<typename T, typename I, class names, bool HD, bool is_function, bool is_deltas, bool is_coords, class ... f_args>
template<bool UPDATE>
grid<T,I,names,HD>  stencil_func<T,I,names,HD,is_function,is_deltas,is_coords,f_args...>::operator*(grid<T,I,names,HD>& _in) const {
    assert(this->props.strides == _in.props.strides);
    grid<T,I,names,HD> output(_in); //should be a shallow copy;
    run_on_grid(_in, output);
    if constexpr (UPDATE) {
        output.update_boundaries();
    }
    return output;
}

template<typename T, typename I, class names, bool HD, bool is_function, bool is_deltas, bool is_coords, class ... f_args>
template<bool is_upper, class new_names, bool new_HD>
typename decide_stencil_func_type<new_names,T,I,is_function,is_deltas,is_coords,HD&&new_HD,stencil_func>::type stencil_func<T,I,names,HD,is_function,is_deltas,is_coords,f_args...>::expand_with(const std::array<I,(new_names::size-names::size)>& array, const std::array<T,(new_names::size-names::size)>& _new_deltas) const {
    const static size_t new_dim = new_names::size;
    const static size_t d = new_dim-names::size; (void) d;
    using new_type = typename  decide_stencil_func_type<new_names,T,I,is_function,is_deltas,is_coords,HD&&new_HD,stencil_func>::type;

    // expand positions
    std::vector<std::array<I,new_dim>> new_positions(this->positions.size());
    auto iter_new=new_positions.begin();
    auto iter_old=this->positions.cbegin();
    for(;iter_old!=this->positions.cend(); ++iter_new, ++iter_old){
        if constexpr (is_upper){
            std::copy(iter_old->cbegin(), iter_old->cend(), iter_new->begin());
            std::copy(array.cbegin(), array.cend(), iter_new->begin()+dim);
        } else {
            std::copy(array.cbegin(), array.cend(), iter_new->begin());
            std::copy(iter_old->cbegin(), iter_old->cend(), iter_new->begin()+d);
        }
    }

    // expand deltas
    std::array<T,new_dim> new_deltas;
    if constexpr (is_upper){
        std::copy(this->deltas.cbegin(), this->deltas.cend(), new_deltas.begin());
        std::copy(_new_deltas.cbegin(), _new_deltas.cend(), new_deltas.begin()+dim);
    } else {
        std::copy(_new_deltas.cbegin(), _new_deltas.cend(), new_deltas.begin());
        std::copy(this->deltas.cbegin(), this->deltas.cend(), new_deltas.begin()+d);
    }

    // expand substencils
    std::vector<new_type> new_stencils;
    new_stencils.reserve(stencils.size());
    for(auto iter=stencils.cbegin(); iter!=stencils.cend(); ++iter){
        new_stencils.push_back(iter->template expand_with<is_upper,new_names,HD&&new_HD>(array,_new_deltas));
    }

    return new_type(this->factor, this->coefficients, new_positions,
                    new_deltas, this->stencil_coefficients, new_stencils, expand_function<is_upper,names::size,new_names::size,T,I,f_args...>::get(this->std_function),this->function_descriptor
    );

    //TODO: check and finish higher stenfil func kron
}


template<typename T, typename I, class names, bool HD, bool is_function, bool is_deltas, bool is_coords, class ... f_args>
std::vector<I> stencil_func<T,I,names,HD,is_function,is_deltas,is_coords,f_args...>::uniq_pos() const {
    std::vector<I> out = this->linear_positions;
    for(auto& el: stencils){
        for(auto& p: el.uniq_pos()){
            for(auto& o: out){
                if(o==p){
                    goto lend;
                }
            }
            out.push_back(p);
            lend: ;
        }
    }
    return out;
}

#endif
