#ifndef stencil_func_add_hpp
#define stencil_func_add_hpp

#include "stencil_func.h"
#include "properties.h"

// Stencil addition
template<typename T, typename I, class names, bool HD, bool is_function, bool is_deltas, bool is_coords, class ... f_args>
inline stencil_func<T,I,names,HD,is_function,is_deltas,is_coords,f_args...> add_stencils(const stencil_func<T,I,names,HD,is_function,is_deltas,is_coords,f_args...>& _s1, const stencil_func<T,I,names,HD,is_function,is_deltas,is_coords,f_args...>& _s2, const T& _f1 = 1., const T& _f2 = 1.){
    // check whether the stencils can be mixed.
    if((_s1.deltas != _s2.deltas) || !(_s1.props == _s2.props)){ std::logic_error("The stencils' deltas do not match.");}

    //NOTE: under normal circumstances this should not be necessary. As functions are not comparable, one might need to use an additional member variable to assertain, whether this is real or not.
//     if(_s1.std_function != function_deflt<names,T,I,f_args...> || _s2.std_function != function_deflt<names,T,I,f_args...>){
//         std::invalid_argument("Only stencils with the default pass-through functional can be sumed up.");
//     }

    std::vector<std::array<I,names::size>> tmp_positions;
    tmp_positions.reserve(_s1.positions.size());
    std::vector<T> tmp_coefficients;
    tmp_coefficients.reserve(_s1.positions.size());

    //     clean and copy _s1
    add_uniq_stencil_points<T,I>(tmp_coefficients, tmp_positions, _s1.coefficients, _s1.positions,_f1);
    add_uniq_stencil_points<T,I>(tmp_coefficients, tmp_positions, _s2.coefficients, _s2.positions,_f2*_s2.factor/_s1.factor);

    // reset the global factor to one if it is an empty stencil
    T tmp_factor = tmp_coefficients.size() == 0 ? 1. : _s1.factor;

    // unite stencil lists
    std::vector<T> tmp_stencil_coefficients;
    tmp_stencil_coefficients.reserve(_s1.stencil_coefficients.size());
    std::vector<stencil_func<T,I,names,HD,is_function,is_deltas,is_coords,f_args...>> tmp_stencils;
    tmp_stencils.reserve(_s1.stencils.size());
    add_uniq_stencil_points<T,I>(tmp_stencil_coefficients, tmp_stencils, _s1.stencil_coefficients, _s1.stencils, _f1);
    add_uniq_stencil_points<T,I>(tmp_stencil_coefficients, tmp_stencils, _s2.stencil_coefficients, _s2.stencils, _f2*_s2.factor/_s1.factor);

    return stencil_func<T,I,names,HD,is_function,is_deltas,is_coords,f_args...>(tmp_factor, tmp_coefficients, tmp_positions, _s1.deltas, tmp_stencil_coefficients, tmp_stencils, _s1.std_function);
}


template<typename T, typename I, class names, bool HD,
        bool is_f_1=false, bool is_d_1=false, bool is_c_1=false, class ... f_args_1,
        bool is_f_2=false, bool is_d_2=false, bool is_c_2=false, class ... f_args_2>
        auto operator+(const stencil_func<T,I,names,HD,is_f_1,is_d_1,is_c_1,f_args_1...>& _s1, const stencil_func<T,I,names,HD,is_f_2,is_d_2,is_c_2,f_args_2...>& _s2){
            if constexpr (std::is_same<stencil_func<T,I,names,HD,is_f_1,is_d_1,is_c_1,f_args_1...>, stencil_func<T,I,names,HD,is_f_2,is_d_2,is_c_2,f_args_2...>>::value){
        return add_stencils(_s1, _s2);
    } else {
        using out_type = typename decide_stencil_func_type<names,T,I,is_f_1||is_f_2,is_d_1||is_d_2,is_c_1||is_c_2,HD, stencil_func>::type;
        return add_stencils((out_type) _s1, (out_type) _s2);
    }
}
template<typename T, typename I, class names, bool HD,
        bool is_f_1=false, bool is_d_1=false, bool is_c_1=false, class ... f_args_1,
        bool is_f_2=false, bool is_d_2=false, bool is_c_2=false, class ... f_args_2>
        auto operator-(const stencil_func<T,I,names,HD,is_f_1,is_d_1,is_c_1,f_args_1...>& _s1, const stencil_func<T,I,names,HD,is_f_2,is_d_2,is_c_2,f_args_2...>& _s2){
            if constexpr (std::is_same<stencil_func<T,I,names,HD,is_f_1,is_d_1,is_c_1,f_args_1...>, stencil_func<T,I,names,HD,is_f_2,is_d_2,is_c_2,f_args_2...>>::value){
        return add_stencils(_s1, _s2, 1., -1.);
    } else {
        using out_type = typename decide_stencil_func_type<names,T,I,is_f_1||is_f_2,is_d_1||is_d_2,is_c_1||is_c_2,HD, stencil_func>::type;
        return add_stencils((out_type) _s1, (out_type) _s2, 1., -1.);
    }
}

// template<class c1, class c2>
// c1 operator+=(c1& _s1, const c2& _s2){
//     if constexpr (std::is_same<c1, c2>::value){
//         _s1 = add_stencils(_s1, _s2);
//     } else {
//         _s1 = add_stencils(_s1, (c1) _s2);
//     }
//     return _s1;
// }
// template<class c1, class c2>
// c1 operator-=(c1& _s1, const c2& _s2){
//     if constexpr (std::is_same<c1, c2>::value){
//         _s1 = add_stencils(_s1, _s2, 1., -1.);
//     } else {
//         _s1 = add_stencils(_s1, (c1) _s2, 1., -1.);
//     }
//     return _s1;
// }
#endif
