#pragma once

#include "stencil_func.h"


template<typename T, typename I, class names, bool HD, bool is_function, bool is_deltas, bool is_coords, class ... f_args>
csr<T,size_t> stencil_func<T,I,names,HD,is_function,is_deltas,is_coords,f_args...>::to_csr( grid<T,I,names,HD>& grid) const{
    if (is_function||is_deltas||is_coords) {
        debug::msg1 << "WARNING: The programmer has to check for himself that all the used functions are linear in nature! This cannot be done in code." << std::endl;
    }

    csr<T,size_t> output;
    auto itr = grid.template begin_intern<is_deltas&&!HD,is_coords>(grid.props.zeroes,grid.props.sizes);
    output.IA.resize(itr.area()+1);
    output.IA[0]=0;

    auto iIA = output.IA.begin()+1;

    size_t tmp;
    size_t counter=0; // this counter basically is the column position of the diagonal
    std::vector<size_t> positions;
    std::vector<I> Ipositions;
    std::vector<T> coefficients;
    auto i_pos = positions.begin();
    auto i_ipos = Ipositions.begin();

//     debug::msg2 << "pos" << std::endl;

    if constexpr (!(is_deltas&&!HD) && !is_coords){
//     debug::msg2 << "pos0a" << std::endl;
        tmp = to_csr_line<I>(coefficients,Ipositions,0);
        positions.resize(Ipositions.size());
    } else {
//     debug::msg2 << "pos0b" << std::endl;
        tmp = uniq_pos().size();
    }
//     debug::msg2 << "pos1 " << tmp << std::endl;
    // the size estimates are done with the maximums
    output.A.reserve(itr.area()*tmp);
    output.JA.reserve(itr.area()*tmp);

    tmp=0;
    for(; itr.within(); ++counter, ++itr, ++iIA){
        if (itr.is_boundary()) {
            *iIA=tmp+1;
            output.A.push_back(1);
            output.JA.push_back(counter);
        } else {
            if constexpr ((is_deltas && !HD) && is_coords) {
                *iIA = tmp + to_csr_line<size_t>(coefficients,positions,counter,itr.deltas,itr.coords);
            } else if constexpr (!is_deltas && is_coords) {
                *iIA=tmp + to_csr_line<size_t>(coefficients,positions,counter,itr.coords);
            } else if constexpr ((is_deltas && !HD) && !is_coords) {
                *iIA=tmp + to_csr_line<size_t>(coefficients,positions,counter,itr.deltas);
            } else {
                *iIA=tmp + to_csr_line<size_t>(coefficients,positions,counter);
                assert(*iIA > tmp);
                i_pos=positions.begin();
                i_ipos=Ipositions.begin();
                for(; i_ipos!=Ipositions.cend(); ++i_ipos, ++i_pos){
                    *i_pos = (size_t) (*i_ipos+counter);
                }
            }
            append_vector(output.A, coefficients);
            append_vector(output.JA, positions);
        }
        tmp = *iIA;
    }

    output.A.shrink_to_fit();
    output.JA.shrink_to_fit();

    return output;
}

template<typename T, typename I, class names, bool HD, bool is_function, bool is_deltas, bool is_coords, class ... f_args>
template<typename Ipos>
size_t  stencil_func<T,I,names,HD,is_function,is_deltas,is_coords,f_args...>::to_csr_line(std::vector<T>& _coefficients, std::vector<Ipos>& _positions, Ipos offset, const f_args& ... args) const {
    _coefficients.resize(0);
    _positions.resize(0);

    T f_val=this->factor;
    if constexpr (is_function) {
        f_val *= this->factor * std_function(1., args...);
    }

    std::vector<T> Ttmp;
    std::vector<Ipos> Itmp;
    auto iter_st = stencils.cbegin();
    auto iter_stc = stencil_coefficients.cbegin();
    for(; iter_st!=stencils.cend(); ++iter_st, ++iter_stc){
        Ttmp.resize(0);
        Itmp.resize(0);
        iter_st->template to_csr_line<Ipos>(Ttmp, Itmp, offset, args...);
        add_uniq_stencil_points<T,I,Ipos>(_coefficients,_positions,Ttmp,Itmp,f_val* *iter_stc);
    }

    Itmp.resize(this->positions.size());
    auto ilp = Itmp.begin();
    auto ip = this->linear_positions.cbegin();
    for(; ip!=this->linear_positions.cend(); ++ip, ++ilp){
        *ilp = (Ipos) (*ip + offset);
    }

    if (!Itmp.empty()){
        add_uniq_stencil_points<T,I,Ipos>(_coefficients,_positions,this->coefficients,Itmp,f_val );
    }

    return _coefficients.size();
}
