#pragma once

#include "stencil_func.h"


template<typename T, typename I, class names, bool HD, bool is_function, bool is_deltas, bool is_coords, class ... f_args>
template<bool cache>
grid<T,I,names,HD> stencil_func<T,I,names,HD,is_function,is_deltas,is_coords,f_args...>::direct_inv(grid<T,I,names,HD> output) {
    debug::msg1 << "Whether the direct inversion delivers correct results also depends on the chosen boundary conditions." << std::endl;
#warning "Whether the direct inversion delivers correct results also depends on the chosen boundary conditions."
    if constexpr (cache) {
        assert(this->props == output.props);
        if (inversion_cache_down.IA.back()==0 && inversion_cache_up.IA.back()==0){
            generate_inversion_cache(output);
        }
        invert_with_cache(output);
    } else {
        invert_no_cache(output);
    }
    return output;
}

template<typename T, typename I, class names, bool HD, bool is_function, bool is_deltas, bool is_coords, class ... f_args>
template<bool BOUNDARIES, typename Ipos, typename ITER>
inline bool stencil_func<T,I,names,HD,is_function,is_deltas,is_coords,f_args...>::get_row(std::vector<Ipos>& _positions, std::vector<T>& _values, const Ipos& counter, const ITER& itr){

    if constexpr (BOUNDARIES){
        if (itr.is_boundary()) {
            _positions = {counter};
            _values = {1};
            return true;
        }
    }

    if constexpr ((is_deltas && !HD) && is_coords) {
        to_csr_line<Ipos>(_values,_positions,counter,itr.deltas,itr.coords);
    } else if constexpr (!is_deltas && is_coords) {
        to_csr_line<Ipos>(_values,_positions,counter,itr.coords);
    } else if constexpr ((is_deltas && !HD) && !is_coords) {
        to_csr_line<Ipos>(_values,_positions,counter,itr.deltas);
    } else {
        to_csr_line<Ipos>(_values,_positions,counter);
        //NOTE: here it would be possible to cache the row:
        //         auto i_pos=_positions.begin();
        //         for(auto i_ipos=I_positions.cbegin(); i_ipos!=I_positions.cend(); ++i_ipos, ++i_pos) {
        //             *i_pos = (Ipos) (*i_ipos+counter);
        //         }
    }
    return false;
}


template<typename T, typename I, class names, bool HD, bool is_function, bool is_deltas, bool is_coords, class ... f_args>
size_t stencil_func<T,I,names,HD,is_function,is_deltas,is_coords,f_args...>::generate_inversion_cache(grid<T,I,names,HD>& output) {
    if (is_function||is_deltas||is_coords) {
        debug::msg1 << "WARNING: The programmer has to check for himself that all the used functions are linear in nature! This cannot be done in code." << std::endl;
    }

    using Ipos = size_t;
    auto itr = output.template begin_intern<is_deltas&&!HD,is_coords>(output.props.zeroes, output.props.sizes);

    size_t counter=0;
    Ipos delta;

    T tmp_coeffi;
    T factor=1.;
    std::vector<std::vector<Ipos>> pos_cache(1);
    std::vector<std::vector<T>> coe_cache(1);
    std::vector<bool> bndry(1);
    bool diagonal_entry = true;

    std::vector<Ipos> JA_down_tmp;
    std::vector<T> A_down_tmp;


    if constexpr (!(is_deltas&&!HD) || !is_coords) {
//         positions.resize(Ipositions.size());
        counter = to_csr_line<Ipos>(coe_cache.front(),pos_cache.front(),0);
    } else {
        counter = uniq_pos().size();
    }
    // the size estimates are done with the maximums
    inversion_cache_up = csr<T,Ipos>(itr.area(), itr.area()*counter);
    inversion_cache_down = csr<T,Ipos>(itr.area(), itr.area()*counter);


    counter=0;
    coe_cache.resize(0);
    pos_cache.resize(0);
    bndry.resize(0);
    for(; itr.within(); ++counter, ++itr, diagonal_entry = true){
        factor=1.;
        if (coe_cache.size() == 0){
            coe_cache.push_back({});
            pos_cache.push_back({});
            bndry.push_back(get_row<true>(pos_cache.back(), coe_cache.back(), counter, itr));
        }
        if (bndry[0]) {
            inversion_cache_down.append_row(std::vector<Ipos>({counter}),std::vector<T>({1.}));
        } else {
            //////// inversion lower inversion part
            A_down_tmp.resize(0);
            JA_down_tmp.resize(0);

            repeat: ;
            for(; pos_cache.front().size()>0 && ((I)pos_cache.front().front()-(I)counter)<0;) {
                delta = pos_cache.front().front();
                assert(delta>=0);
                assert(delta<(I)itr.area());
                assert(delta<(I)counter);
                assert((inversion_cache_up.IA[delta]-inversion_cache_up.IA[delta-1])>0);

                // eliminate lower triangular elements of the matrix
                tmp_coeffi = coe_cache.front().front();
                pos_cache.front().erase(pos_cache.front().begin());
                coe_cache.front().erase(coe_cache.front().begin());

                if(! is_about_zero(tmp_coeffi)) {
                    add_uniq_stencil_points<T,I,Ipos>(A_down_tmp, JA_down_tmp, {tmp_coeffi}, {delta}, 1.);
                    add_uniq_stencil_points<T,I,Ipos>(coe_cache.front(), pos_cache.front(),
                                inversion_cache_up.A.cbegin()+inversion_cache_up.IA[delta]+1, inversion_cache_up.A.cbegin()+inversion_cache_up.IA[delta+1],
                                inversion_cache_up.JA.cbegin()+inversion_cache_up.IA[delta]+1, inversion_cache_up.JA.cbegin()+inversion_cache_up.IA[delta+1],
                                -tmp_coeffi
                        );
                }
            }

            // divide line by diagonal coefficient
            if (pos_cache.front().size()>0 && pos_cache.front().front()==counter && !is_about_zero(coe_cache.front().front())){
                JA_down_tmp.push_back(pos_cache.front().front()*factor);
                A_down_tmp.push_back(coe_cache.front().front());
                for(auto liter = coe_cache.front().rbegin(); liter!=coe_cache.front().rend(); ++liter){
                    *liter/=coe_cache.front().front()*factor;
                }
            } else if (diagonal_entry) {
                diagonal_entry = false;
                auto tmp_itr = itr;
                ++tmp_itr;
                auto tmp_counter = counter+1;

                auto pos_iter = pos_cache.cbegin()+1;
                auto coe_iter = coe_cache.cbegin()+1;
                auto bndry_iter = bndry.cbegin()+1;
                for(;tmp_itr.within(); ++tmp_itr, ++tmp_counter, ++pos_iter, ++coe_iter, ++bndry_iter){
                    if (bndry_iter==bndry.cend()){
                        coe_cache.push_back({});
                        pos_cache.push_back({});
                        bndry.push_back(get_row<true>(pos_cache.back(), coe_cache.back(), tmp_counter, tmp_itr));
                        bndry_iter=bndry.cend()-1;
                        pos_iter=pos_cache.cend()-1;
                        coe_iter=coe_cache.cend()-1;
                    }
                    if (!*bndry_iter){
                        auto rslt = std::find(pos_iter->cbegin(),pos_iter->cend(),counter);
                        if ( rslt != pos_iter->cend()){
                            add_uniq_stencil_points<T,I,Ipos>(coe_cache.front(),pos_cache.front(),*coe_iter,*pos_iter,1./(*coe_iter)[rslt-pos_iter->begin()]);
                            factor=2.;

                            goto repeat;
                        }
                    }
                }
                goto fail;

            } else {
                fail: ;
                // At this point we basically have an upper triangular matrix. Hence, we can make statements on the invertibility.
                assert(false && "This stencil is not invertible. The current algorithm requires a diagonal element.");
            }

            //////// end lower inversion part

            //                 assert(coe_cache.front().size()>0);
            inversion_cache_down.append_row(JA_down_tmp,A_down_tmp);
        }
        inversion_cache_up.append_row(pos_cache.front(), coe_cache.front());
        coe_cache.erase(coe_cache.begin());
        pos_cache.erase(pos_cache.begin());
        bndry.erase(bndry.begin());

    }
    inversion_cache_up.shrink_to_fit();
    inversion_cache_down.shrink_to_fit();

    return counter;
}

/////////////// STOP STOP STOP STOP
template<typename T, typename I, class names, bool HD, bool is_function, bool is_deltas, bool is_coords, class ... f_args>
inline void stencil_func<T,I,names,HD,is_function,is_deltas,is_coords,f_args...>::invert_no_cache(grid<T,I,names,HD>& output) {
    if (is_function||is_deltas||is_coords) {
        debug::msg1 << "WARNING: The programmer has to check for himself that all the used functions are linear in nature! This cannot be done in code." << std::endl;
    }

    using Ipos = T*;
    auto itr = output.template begin_intern<is_deltas&&!HD,is_coords>(output.props.zeroes,output.props.sizes);

    size_t tmp;
    size_t counter=0;
    I delta;
    T tmp_coeffi;

    std::vector<Ipos> positions;
    std::vector<T> coefficients;
    std::vector<Ipos> JA_down_tmp;
    std::vector<T> A_down_tmp;

    if constexpr (!(is_deltas&&!HD) || !is_coords) {
        tmp = to_csr_line<Ipos>(coefficients,positions,0);
//         positions.resize(Ipositions.size());
    } else {
        tmp = uniq_pos().size();
    }
    // the size estimates are done with the maximums
    csr<T,T*> memory(itr.area(), itr.area()*tmp);


    tmp=0;
    for(; itr.within(); ++counter, ++itr){
        if (itr.is_boundary()) {
            memory.append_row(std::vector<T*>({itr.ptr}),std::vector<T>({1.}));

            //NOTE: as the boundaries are basically unit blocks, these elements may be rationalized away. But, as the boundaries should be small compared to the inner domain, it might not be really necessary.
//                 inversion_cache_up.append_empty_row();
        } else {
            get_row(positions, coefficients, itr.ptr, itr);

            //////// inversion lower inversion part
            A_down_tmp.resize(0);
            JA_down_tmp.resize(0);
            for(; positions.size()>0 && (positions.front()-itr.ptr)<0;) {
                delta = positions.front() - output.storage.data();
                assert(delta>=0);
                assert(delta<(I)itr.area());
                assert(delta<(I)counter);
                assert((memory.IA[delta]-memory.IA[delta-1])>0);

                // eliminate lower triangular elements of the matrix
                tmp_coeffi = coefficients.front();
                *(itr.ptr) -= tmp_coeffi * *(positions.front());
                A_down_tmp.push_back(tmp_coeffi);
                JA_down_tmp.push_back(positions.front());
                positions.erase(positions.begin());
                coefficients.erase(coefficients.begin());

                add_uniq_stencil_points<T,I,Ipos>(coefficients, positions,
                            memory.A.cbegin()+memory.IA[delta]+1, memory.A.cbegin()+memory.IA[delta+1],
                            memory.JA.cbegin()+memory.IA[delta]+1, memory.JA.cbegin()+memory.IA[delta+1],
                            -tmp_coeffi
                    );
            }


            // divide line by diagonal coefficient
            if (positions.size()>0 && (positions.front()-itr.ptr)==0){
                *(itr.ptr)/=coefficients.front();
                for(auto liter = coefficients.rbegin(); liter!=coefficients.rend()-1; ++liter){
                    *liter/=coefficients.front();
                }
                JA_down_tmp.push_back(positions.front());
                A_down_tmp.push_back(coefficients.front());
            } else {
                // At this point we basically have an upper triangular matrix. Hence, we can make statements on the invertibility.
                assert(false && "This stencil is not invertible. The current algorithm requires a diagonal element. The default cached version might in same cases work, because it is more complete.");
            }

            //////// end lower inversion part

//                 assert(coefficients.size()>0);
            memory.append_row(positions, coefficients);
        }
    }

        memory.shrink_to_fit();


    ////////// begin upper inversion part
    assert(memory.IA.size()-1 == itr.area());
    auto iter = memory.IA.crbegin()+1;
    itr.reset_end();
    size_t i;
    tmp = memory.IA.back();
    for(; iter!=memory.IA.crend() && itr.within(); ++iter, --itr){
        if (!itr.is_boundary()) {
            for (i=tmp-1; i>*iter; --i) {
                *(memory.JA[*iter]) -= (memory.A[i])* *(memory.JA[i]);
            }
        }
        tmp = *iter;
    }
    ////////// end upper inversion part

}


//////////stop stop stop
template<typename T, typename I, class names, bool HD, bool is_function, bool is_deltas, bool is_coords, class ... f_args>
inline void stencil_func<T,I,names,HD,is_function,is_deltas,is_coords,f_args...>::invert_with_cache(grid<T,I,names,HD>& output) {
    if (is_function||is_deltas||is_coords) {
        debug::msg1 << "WARNING: The programmer has to check for himself that all the used functions are linear in nature! This cannot be done in code." << std::endl;
    }

    using Ipos = T*;
    Ipos offset = output.storage.data();

    auto itr = output.template begin_intern<is_deltas&&!HD,is_coords>(output.props.zeroes,output.props.sizes);

    size_t tmp;

    std::vector<Ipos> positions;
    std::vector<T> coefficients;
    std::vector<Ipos> JA_down_tmp;
    std::vector<T> A_down_tmp;

    tmp = inversion_cache_down.IA.front();
    size_t i;
    auto iter_d = inversion_cache_down.IA.cbegin()+1;
    for(; iter_d!=inversion_cache_down.IA.cend(); ++iter_d){
        for(i=tmp; i<*iter_d-1; ++i){
            *(offset+inversion_cache_down.JA[*iter_d-1]) -= (inversion_cache_down.A[i])* *(offset+inversion_cache_down.JA[i]);
        }
        *(inversion_cache_down.JA[*iter_d-1]+offset)/=inversion_cache_down.A[i];
        tmp=*iter_d;
    }


    ////////// begin upper inversion part
    assert(inversion_cache_up.IA.size()-1 == itr.area());
    auto iter = inversion_cache_up.IA.crbegin()+1;
    itr.reset_end();
    tmp = inversion_cache_up.IA.back();
    for(; iter!=inversion_cache_up.IA.crend() && itr.within(); ++iter, --itr){
        if (!itr.is_boundary()) {
            for (i=tmp-1; i>*iter; --i) {
                *(inversion_cache_up.JA[*iter]+offset) -= (inversion_cache_up.A[i])* *(inversion_cache_up.JA[i]+offset);
            }
        }
        tmp = *iter;
    }
    ////////// end upper inversion part

}

