#ifndef stencil_func_multiply_hpp
#define stencil_func_multiply_hpp

#include "stencil_func.h"
#include "stencil_func_helpers.h" //TODO split into declaration and definition
#include <algorithm>
#include "helpers.h"

template<class names, typename T, typename T2, typename I, bool is_function, bool is_deltas, bool is_coords, bool HD, class ... f_args>
stencil_func<T,I,names,HD,is_function,is_deltas,is_coords,f_args...> operator*(const stencil_func<T,I,names,HD,is_function,is_deltas,is_coords,f_args...>& _stencil_in, T2 _factor){
    static_assert(std::is_arithmetic<T2>::value);
    if(is_about_zero<T2>(_factor)){
        debug::msg3 << "factor is about zero." << std::endl;
        return stencil_func<T,I,names,HD,is_function,is_deltas,is_coords,f_args...>();
    }
    return stencil_func<T,I,names,HD,is_function,is_deltas,is_coords,f_args...>(_stencil_in.factor * _factor, _stencil_in.coefficients, _stencil_in.positions, _stencil_in.deltas,  _stencil_in.stencil_coefficients, _stencil_in.stencils, _stencil_in.std_function);
}

template<class names, typename T, typename T2, typename I, bool is_function, bool is_deltas, bool is_coords, bool HD, class ... f_args>
stencil_func<T,I,names,HD,is_function,is_deltas,is_coords,f_args...> operator*(const T2 _factor, const stencil_func<T,I,names,HD,is_function,is_deltas,is_coords,f_args...>& _stencil_in){
    return operator*(_stencil_in, (T) _factor);
}

template<class names, typename T, typename T2, typename I, bool is_function, bool is_deltas, bool is_coords, bool HD, class ... f_args>
stencil_func<T,I,names,HD,is_function,is_deltas,is_coords,f_args...> operator*=(stencil_func<T,I,names,HD,is_function,is_deltas,is_coords,f_args...>& _stencil_in, const T2 _factor){
    _stencil_in = operator*(_stencil_in, (T) _factor);
    return _stencil_in;
}

template<class names, typename T, typename T2, typename I, bool is_function, bool is_deltas, bool is_coords, bool HD, class ... f_args>
stencil_func<T,I,names,HD,is_function,is_deltas,is_coords,f_args...> operator/(const stencil_func<T,I,names,HD,is_function,is_deltas,is_coords,f_args...>& _stencil_in, const T2 _factor){
    if(is_about_zero<T2>(_factor)){
        std::invalid_argument("Division by zero in stencil_func::operator/");
    }
    return stencil_func<T,I,names,HD,is_function,is_deltas,is_coords,f_args...>(_stencil_in.factor / (T) _factor, _stencil_in.coefficients, _stencil_in.positions, _stencil_in.deltas, _stencil_in.stencil_coefficients, _stencil_in.stencils, _stencil_in.std_function);
}

template<class names, typename T, typename T2, typename I, bool is_function, bool is_deltas, bool is_coords, bool HD, class ... f_args>
stencil_func<T,I,names,HD,is_function,is_deltas,is_coords,f_args...> operator/=(stencil_func<T,I,names,HD,is_function,is_deltas,is_coords,f_args...>& _stencil_in, const T2 _factor){
    _stencil_in = operator/(_stencil_in, (T) _factor);
    return _stencil_in;
}



template<typename T, typename I, class names, bool HD, bool is_function, bool is_deltas, bool is_coords, class ... f_args>
template<class names_new, bool HD_new, bool is_function_new, bool is_deltas_new, bool is_coords_new, class ... f_args_new>
auto stencil_func<T,I,names,HD,is_function,is_deltas,is_coords,f_args...>::kron(const stencil_func<T,I,names_new,HD_new,is_function_new,is_deltas_new,is_coords_new,f_args_new...>& _in){
    using new_names = typename merge_<char,names,names_new>::list;
    static_assert(new_names::is_uniq(),"Only disjunct dimensions allowed!");

    constexpr static bool hd_out = HD&&HD_new;

    properties<T,I,new_names> props_new(hd_out, array_concat(this->props.gc_lower, _in.props.gc_lower), array_concat(this->props.points, _in.props.points), array_concat(this->props.gc_upper, _in.props.gc_upper));


    using new_type = typename decide_stencil_func_type<new_names,T,I,is_function_new||is_function,is_deltas_new||is_deltas,is_coords_new||is_coords,hd_out,stencil_func>::type;

    assert (_in.stencils.size()==0);


    // stencil_proto * stencil_proto
    typename std::vector<std::array<I,new_names::size>> pos_full;
    typename std::vector<T> coeffi_full;
    auto i_pf = this->positions.cbegin();
    auto i_c = this->coefficients.cbegin();
    auto i_pf2 = _in.positions.cbegin();
    auto i_c2 = _in.coefficients.cbegin();

    typename std::array<T,new_names::size> tmpT;

    for(; i_pf != this->positions.cend(); ++i_pf, ++i_c){
        i_pf2 = _in.positions.cbegin();
        i_c2 = _in.coefficients.cbegin();
        for(; i_pf2 != _in.positions.cend(); ++i_pf2, ++i_c2){
            pos_full.push_back(array_concat(*i_pf,*i_pf2));
            coeffi_full.push_back(*i_c * *i_c2);
        }
    }

    std::merge(deltas.cbegin(),deltas.cend(),_in.deltas.cbegin(),_in.deltas.cend(),tmpT.begin());

//     new_names::debug();
//     new_strides::debug();
//     debug::msg2 << new_strides::size << std::endl;

    //positions * stencils
    std::vector<T> stencil_coefficients_new;
    std::vector<new_type> stencils_new;
    for(i_pf=this->positions.cbegin(), i_c=this->coefficients.cbegin(); i_pf != this->positions.cend(); ++i_pf, ++i_c){
        auto i_s2 = _in.stencils.cbegin();
        auto i_c2 = _in.stencil_coefficients.cbegin();
        for(; i_s2 != _in.stencils.cend(); ++i_s2, ++i_c2){
            stencil_coefficients_new.push_back(*i_c * *i_c2);
            stencils_new.push_back(i_s2->template expand_with<false,new_names,hd_out>(*i_pf,_in.deltas));
        }
    }

    // stencils * positions
    for(i_pf=_in.positions.cbegin(), i_c=_in.coefficients.cbegin(); i_pf != _in.positions.cend(); ++i_pf, ++i_c){
        auto i_s2 = this->stencils.cbegin();
        auto i_c2 = this->stencil_coefficients.cbegin();
        for(; i_s2 != this->stencils.cend(); ++i_s2, ++i_c2){
            stencil_coefficients_new.push_back(*i_c * *i_c2);
            stencils_new.push_back(i_s2->template expand_with<true,new_names,hd_out>(*i_pf,_in.deltas));
        }
    }


    // stencils * stencils
    assert(((this->stencils.size()>=0&&_in.stencils.size()==0)||(this->stencils.size()==0&&_in.stencils.size()>=0))&&"Only one of the stencils may be functionalized!");

    return new_type(this->factor * (T)_in.factor, coeffi_full, pos_full, tmpT, stencil_coefficients_new, stencils_new);
}

template<typename T, typename I, class names, bool HD, bool is_function, bool is_deltas, bool is_coords, class ... f_args>
auto stencil_product(const stencil_func<T,I,names,HD,is_function,is_deltas,is_coords,f_args...>& _s1, const stencil_func<T,I,names,HD,is_function,is_deltas,is_coords,f_args...>& _s2){
    using out_type = stencil_func<T,I,names,HD,is_function,is_deltas,is_coords,f_args...>;

    typename std::vector<std::array<I,names::size>> pos_full;
    typename std::vector<T> coeffi_full;
    typename std::vector<std::array<I,names::size>> pos_final;
    typename std::vector<T> coeffi_final;
    auto i_pf = _s1.positions.cbegin();
    auto i_c = _s1.coefficients.cbegin();
    auto i_pf2 = _s2.positions.cbegin();
    auto i_c2 = _s2.coefficients.cbegin();

    // stencil points
    for(; i_pf != _s1.positions.cend(); ++i_pf, ++i_c){
        i_pf2 = _s2.positions.cbegin();
        i_c2 = _s2.coefficients.cbegin();
        for(; i_pf2 != _s2.positions.cend(); ++i_pf2, ++i_c2){
            pos_full.push_back(*i_pf+*i_pf2);
            coeffi_full.push_back(*i_c * *i_c2);
        }
    }
//     debug::msg2 << coeffi_full << pos_full << std::endl;
    add_uniq_stencil_points<T,I>(coeffi_final, pos_final, coeffi_full, pos_full);

    if (!is_function && !is_deltas && !is_coords && (_s1.stencils.size() + _s2.stencils.size() == 0)){
        return out_type(_s1.factor * _s2.factor, coeffi_final, pos_final);
    }
    assert(_s1.deltas == _s2.deltas);

    debug::msg1 << "Please remember that the stencil product requires linear stencil functions!" << std::endl;

    // positions * stencils
    i_pf = _s1.positions.cbegin();
    i_c = _s1.coefficients.cbegin();
    std::vector<T> new_st_coef;
    std::vector<out_type> new_st;
    for(; i_pf != _s1.positions.cend(); ++i_pf, ++i_c){
        auto i_st2 = _s2.stencils.cbegin();
        i_c2 = _s2.stencil_coefficients.cbegin();
        for(; i_st2 != _s2.stencils.cend(); ++i_st2, ++i_c2){
            add_uniq_stencil_points<T,I>(new_st_coef,new_st,std::vector<T> {*i_c * *i_c2},std::vector<out_type>{i_st2->shift(*i_pf)});
        }
    }

    auto i_st1 = _s1.stencils.cbegin();
    auto i_stc1 = _s1.stencil_coefficients.cbegin();
    for(; i_st1!=_s1.stencils.cend(); ++i_st1, ++i_stc1){
        // stencils * positions
        add_uniq_stencil_points<T,I>(new_st_coef,new_st,std::vector<T> {*i_stc1},std::vector<out_type>{i_st1->template shift<false>(_s2.coefficients,_s2.positions)});

    // stencils * stencils
        auto itr = _s2.stencils.cbegin();
        auto citr = _s2.stencil_coefficients.cbegin();
        std::vector<out_type> newsubstencils;
        for(; itr!=_s2.stencils.cend(); ++itr, ++citr){
            newsubstencils.push_back(stencil_product(*i_st1,*itr));
        }
        add_uniq_stencil_points<T,I,out_type>(new_st_coef,new_st,_s2.stencil_coefficients,newsubstencils,*i_stc1);
    }


    return stencil_func<T,I,names,HD,is_function,is_deltas,is_coords,f_args...>(_s1.factor * _s2.factor, coeffi_final, pos_final, _s1.deltas, new_st_coef, new_st, _s1.std_function, _s1.function_descriptor);
}

template<typename T, typename I, class names, bool HD,
bool is_f_1=false, bool is_d_1=false, bool is_c_1=false, class ... f_args_1,
bool is_f_2=false, bool is_d_2=false, bool is_c_2=false, class ... f_args_2>
auto operator*(const stencil_func<T,I,names,HD,is_f_1,is_d_1,is_c_1,f_args_1...>& _s1, const stencil_func<T,I,names,HD,is_f_2,is_d_2,is_c_2,f_args_2...>& _s2){
    if constexpr (std::is_same<stencil_func<T,I,names,HD,is_f_1,is_d_1,is_c_1,f_args_1...>, stencil_func<T,I,names,HD,is_f_2,is_d_2,is_c_2,f_args_2...>>::value){
        return stencil_product(_s1, _s2);
    } else {
        using out_type = typename decide_stencil_func_type<names,T,I,is_f_1||is_f_2,is_d_1||is_d_2,is_c_1||is_c_2,HD, stencil_func>::type;
        return stencil_product((out_type) _s1, (out_type) _s2);
    }
}

#endif
