#ifndef stencil_proto_h
#define stencil_proto_h

#include <vector>
#include <cassert>

#include "matrix.h"
#include "liste.h"
#include "helpers.h"
#include "grid.h"
#include "dim.h"
#include "properties.h"

template<typename T, typename I, class names>
struct stencil_proto : public matrix<names,T,I> {
    constexpr static int dim = names::size;

    // constructors

    stencil_proto() :
            matrix<names,T,I>(),
            props(properties<T,I,names>()),
            factor(1.),
            coefficients(std::vector<T>(0)),
            positions(std::vector<std::array<I,dim>>(0)),
            linear_positions(std::vector<I>(0)) {}

    template<bool HD>
    stencil_proto(const grid<T,I,names,HD>& _grid, T _factor, const std::vector<T>& _coefficients, const std::vector<std::array<I,names::size>>& _positions) :
            matrix<names,T,I>(),
            props(_grid.props),
            factor(_factor),
            coefficients(_coefficients),
            positions(_positions),
            linear_positions(linearize_positions(_positions)) {
                assert(props.size>0);
                if (props.size>1) {
                    check_sanity(_positions);
                }
            }

    stencil_proto(const T& _factor, const std::vector<T>& _coefficients, const std::vector<std::array<I,dim>>& _positions) :
            matrix<names,T,I>(),
            props(properties<T,I,names>()),
            factor(_factor),
            coefficients(_coefficients),
            positions(_positions),
            linear_positions(linearize_positions(_positions)) {
                assert(props.size>0);
                if (props.size>1) {
                    check_sanity(_positions);
                }
            }


    stencil_proto(const stencil_proto<T,I,names>& _stencil) :
            matrix<names,T,I>(),
            props(_stencil.props),
            factor(_stencil.factor),
            coefficients(_stencil.coefficients),
            positions(_stencil.positions),
            linear_positions(_stencil.linear_positions) {}


    // data
    const properties<T,I,names> props;
    const T factor;
    const std::vector<T> coefficients;
    const std::vector<std::array<I,dim>> positions;
    const std::vector<I> linear_positions;

    template<typename ITER = typename std::vector<T>::const_iterator>
    constexpr inline T function(ITER) const;

    stencil_proto<T,I,names>& operator=(const stencil_proto<T,I,names>& _input);

protected:
    std::vector<I> linearize_positions(const std::vector<std::array<I,names::size>>& _positions) ;
    void check_sanity(const std::vector<std::array<I,dim>>&);
};


#include "stencil_proto.hpp"
#endif
