#ifndef stencil_proto_hpp
#define stencil_proto_hpp

#include "stencil_proto.h"

template<typename T, typename I, class names>
std::vector<I> stencil_proto<T,I,names>::linearize_positions(const std::vector<std::array<I,names::size>>& _positions) {
    std::vector<I> output(_positions.size());
    typename std::vector<std::array<I,dim>>::const_iterator iter_positions = _positions.cbegin();
    typename std::vector<I>::iterator iter_linear = output.begin();

    for(;iter_positions != _positions.cend(); ++iter_positions, ++iter_linear){
        *iter_linear = props.template inner_product<false>(*iter_positions);
    }
    return output;
}

template<typename T, typename I, class names>
template<typename ITER>
constexpr inline T stencil_proto<T,I,names>::function(ITER _in) const {
    T output = 0;
    auto iter_coeffi = coefficients.cbegin();
    auto iter_pos = linear_positions.cbegin();
//     debug::msg3 << linear_positions << std::endl;
    for(; iter_coeffi!=coefficients.cend(); ++iter_coeffi, ++iter_pos){
//         debug::msg3 << *iter_pos << ": " << *(_in+*iter_pos) << std::endl;
        output += *iter_coeffi * *(_in+*iter_pos);
    }
    return factor * output;
}

template<typename T, typename I, class names>
void stencil_proto<T,I,names>::check_sanity(const std::vector<std::array<I,dim>>& _input){
    bool val=true;
    for(auto& el: _input){
        val = val && props.template within_domain<false,I>(el);
        if (!val){ break;}
    }
    assert(val && "At least one stencil not within grid bounds.");
}

template<typename T, typename I, class names>
stencil_proto<T,I,names>& stencil_proto<T,I,names>::operator=(const stencil_proto<T,I,names>& _input){
    if(this == &_input){
        return *this;
    }

    new (this) stencil_proto<T,I,names>(_input);
    return *this;
}

#endif
